# Copyright (c) 2013-2016 Cisco Systems, Inc.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import collections
import mock
from oslo_config import cfg
import six

from networking_cisco.backwards_compatibility import cb_constants
from networking_cisco.ml2_drivers.nexus import (
    constants as const)
from networking_cisco.ml2_drivers.nexus import (
    nexus_network_driver)
from networking_cisco.ml2_drivers.nexus import exceptions
from networking_cisco.ml2_drivers.nexus import nexus_db_v2

from networking_cisco.tests.unit.ml2_drivers.nexus import (
    test_cisco_nexus_base)

CONNECT_ERROR = 'Unable to connect to Nexus'


class TestCiscoNexusDeviceConfig(object):

    """Unit tests Config for Cisco ML2 Nexus device driver."""

    test_configs = {
        'test_config1':
            test_cisco_nexus_base.TestCiscoNexusBase.TestConfigObj(
                test_cisco_nexus_base.NEXUS_IP_ADDRESS_1,
                test_cisco_nexus_base.HOST_NAME_1,
                test_cisco_nexus_base.NEXUS_PORT_1,
                test_cisco_nexus_base.INSTANCE_1,
                test_cisco_nexus_base.VLAN_ID_1,
                test_cisco_nexus_base.NO_VXLAN_ID,
                None,
                test_cisco_nexus_base.DEVICE_OWNER_COMPUTE,
                {},
                None,
                test_cisco_nexus_base.NORMAL_VNIC),
        'test_config2':
            test_cisco_nexus_base.TestCiscoNexusBase.TestConfigObj(
                test_cisco_nexus_base.NEXUS_IP_ADDRESS_1,
                test_cisco_nexus_base.HOST_NAME_2,
                test_cisco_nexus_base.NEXUS_PORT_2,
                test_cisco_nexus_base.INSTANCE_2,
                test_cisco_nexus_base.VLAN_ID_2,
                test_cisco_nexus_base.NO_VXLAN_ID,
                None,
                test_cisco_nexus_base.DEVICE_OWNER_COMPUTE,
                {},
                None,
                test_cisco_nexus_base.NORMAL_VNIC),
        'test_config3':
            test_cisco_nexus_base.TestCiscoNexusBase.TestConfigObj(
                test_cisco_nexus_base.NEXUS_IP_ADDRESS_1,
                test_cisco_nexus_base.HOST_NAME_1,
                test_cisco_nexus_base.NEXUS_PORT_1,
                test_cisco_nexus_base.INSTANCE_2,
                test_cisco_nexus_base.VLAN_ID_1,
                test_cisco_nexus_base.NO_VXLAN_ID,
                None,
                test_cisco_nexus_base.DEVICE_OWNER_COMPUTE,
                {},
                None,
                test_cisco_nexus_base.NORMAL_VNIC),
        'test_config4':
            test_cisco_nexus_base.TestCiscoNexusBase.TestConfigObj(
                test_cisco_nexus_base.NEXUS_IP_ADDRESS_8,
                test_cisco_nexus_base.HOST_NAME_4,
                test_cisco_nexus_base.NEXUS_PORT_1,
                test_cisco_nexus_base.INSTANCE_1,
                test_cisco_nexus_base.VLAN_ID_1,
                test_cisco_nexus_base.NO_VXLAN_ID,
                None,
                test_cisco_nexus_base.DEVICE_OWNER_COMPUTE,
                {},
                None,
                test_cisco_nexus_base.NORMAL_VNIC),
        'test_config5':
            test_cisco_nexus_base.TestCiscoNexusBase.TestConfigObj(
                test_cisco_nexus_base.NEXUS_IP_ADDRESS_8,
                test_cisco_nexus_base.HOST_NAME_5,
                test_cisco_nexus_base.NEXUS_PORT_2,
                test_cisco_nexus_base.INSTANCE_2,
                test_cisco_nexus_base.VLAN_ID_1,
                test_cisco_nexus_base.NO_VXLAN_ID,
                None,
                test_cisco_nexus_base.DEVICE_OWNER_COMPUTE,
                {},
                None,
                test_cisco_nexus_base.NORMAL_VNIC),
        'test_config_portchannel':
            test_cisco_nexus_base.TestCiscoNexusBase.TestConfigObj(
                test_cisco_nexus_base.NEXUS_IP_ADDRESS_2,
                test_cisco_nexus_base.HOST_NAME_PC,
                test_cisco_nexus_base.NEXUS_PORTCHANNELS,
                test_cisco_nexus_base.INSTANCE_PC,
                test_cisco_nexus_base.VLAN_ID_PC,
                test_cisco_nexus_base.NO_VXLAN_ID,
                None,
                test_cisco_nexus_base.DEVICE_OWNER_COMPUTE,
                {},
                None,
                test_cisco_nexus_base.NORMAL_VNIC),
        'test_config_dual':
            test_cisco_nexus_base.TestCiscoNexusBase.TestConfigObj(
                test_cisco_nexus_base.NEXUS_IP_ADDRESS_DUAL,
                test_cisco_nexus_base.HOST_NAME_DUAL,
                test_cisco_nexus_base.NEXUS_DUAL,
                test_cisco_nexus_base.INSTANCE_DUAL,
                test_cisco_nexus_base.VLAN_ID_DUAL,
                test_cisco_nexus_base.NO_VXLAN_ID,
                None,
                test_cisco_nexus_base.DEVICE_OWNER_COMPUTE,
                {},
                None,
                test_cisco_nexus_base.NORMAL_VNIC),
        'test_config_dhcp':
            test_cisco_nexus_base.TestCiscoNexusBase.TestConfigObj(
                test_cisco_nexus_base.NEXUS_IP_ADDRESS_1,
                test_cisco_nexus_base.HOST_NAME_1,
                test_cisco_nexus_base.NEXUS_PORT_1,
                test_cisco_nexus_base.INSTANCE_1,
                test_cisco_nexus_base.VLAN_ID_1,
                test_cisco_nexus_base.NO_VXLAN_ID,
                None,
                test_cisco_nexus_base.DEVICE_OWNER_DHCP,
                {},
                None,
                test_cisco_nexus_base.NORMAL_VNIC),
        'test_config_router_ha_intf':
            test_cisco_nexus_base.TestCiscoNexusBase.TestConfigObj(
                test_cisco_nexus_base.NEXUS_IP_ADDRESS_1,
                test_cisco_nexus_base.HOST_NAME_1,
                test_cisco_nexus_base.NEXUS_PORT_1,
                test_cisco_nexus_base.INSTANCE_1,
                test_cisco_nexus_base.VLAN_ID_1,
                test_cisco_nexus_base.NO_VXLAN_ID,
                None,
                test_cisco_nexus_base.DEVICE_OWNER_ROUTER_HA_INTF,
                {},
                None,
                test_cisco_nexus_base.NORMAL_VNIC),
        'test_config_router_intf':
            test_cisco_nexus_base.TestCiscoNexusBase.TestConfigObj(
                test_cisco_nexus_base.NEXUS_IP_ADDRESS_1,
                test_cisco_nexus_base.HOST_NAME_1,
                test_cisco_nexus_base.NEXUS_PORT_1,
                test_cisco_nexus_base.INSTANCE_1,
                test_cisco_nexus_base.VLAN_ID_1,
                test_cisco_nexus_base.NO_VXLAN_ID,
                None,
                test_cisco_nexus_base.DEVICE_OWNER_ROUTER_INTF,
                {},
                None,
                test_cisco_nexus_base.NORMAL_VNIC),
        'test_config_router_gw':
            test_cisco_nexus_base.TestCiscoNexusBase.TestConfigObj(
                test_cisco_nexus_base.NEXUS_IP_ADDRESS_1,
                test_cisco_nexus_base.HOST_NAME_1,
                test_cisco_nexus_base.NEXUS_PORT_1,
                test_cisco_nexus_base.INSTANCE_1,
                test_cisco_nexus_base.VLAN_ID_1,
                test_cisco_nexus_base.NO_VXLAN_ID,
                None,
                test_cisco_nexus_base.DEVICE_OWNER_ROUTER_GW,
                {},
                None,
                test_cisco_nexus_base.NORMAL_VNIC),
        'test_config_portchannel2':
            test_cisco_nexus_base.TestCiscoNexusBase.TestConfigObj(
                test_cisco_nexus_base.NEXUS_IP_ADDRESS_6,
                test_cisco_nexus_base.HOST_NAME_3,
                test_cisco_nexus_base.NEXUS_PORTCHANNELS,
                test_cisco_nexus_base.INSTANCE_PC,
                test_cisco_nexus_base.VLAN_ID_PC,
                test_cisco_nexus_base.NO_VXLAN_ID,
                None,
                test_cisco_nexus_base.DEVICE_OWNER_COMPUTE,
                {},
                None,
                test_cisco_nexus_base.NORMAL_VNIC),
        'test_config_portchannel3':
            test_cisco_nexus_base.TestCiscoNexusBase.TestConfigObj(
                test_cisco_nexus_base.NEXUS_IP_ADDRESS_7,
                test_cisco_nexus_base.HOST_NAME_3,
                test_cisco_nexus_base.NEXUS_PORTCHANNELS,
                test_cisco_nexus_base.INSTANCE_PC,
                test_cisco_nexus_base.VLAN_ID_PC,
                test_cisco_nexus_base.NO_VXLAN_ID,
                None,
                test_cisco_nexus_base.DEVICE_OWNER_COMPUTE,
                {},
                None,
                test_cisco_nexus_base.NORMAL_VNIC),
        'test_config_migrate':
            test_cisco_nexus_base.TestCiscoNexusBase.TestConfigObj(
                test_cisco_nexus_base.NEXUS_IP_ADDRESS_3,
                test_cisco_nexus_base.HOST_NAME_6,
                test_cisco_nexus_base.NEXUS_PORT_2,
                test_cisco_nexus_base.INSTANCE_1,
                test_cisco_nexus_base.VLAN_ID_1,
                test_cisco_nexus_base.NO_VXLAN_ID,
                None,
                test_cisco_nexus_base.DEVICE_OWNER_COMPUTE,
                {},
                None,
                test_cisco_nexus_base.NORMAL_VNIC),
    }

    test_configs = collections.OrderedDict(sorted(test_configs.items()))


class TestCiscoNexusDeviceResults(
    test_cisco_nexus_base.TestCiscoNexusBaseResults):

    """Unit tests driver results for Cisco ML2 Nexus."""

    test_results = {

        'duplicate_add_port_driver_result': (
            [test_cisco_nexus_base.RESULT_ADD_VLAN.format(267),
            test_cisco_nexus_base.RESULT_ADD_INTERFACE.
                format('ethernet', '1\/10', 267)]),

        'duplicate_del_port_driver_result': (
            [test_cisco_nexus_base.RESULT_DEL_INTERFACE.
                format('ethernet', '1\/10', 267),
            test_cisco_nexus_base.RESULT_DEL_VLAN.format(267)]),

        'add_port2_driver_result': (
            [test_cisco_nexus_base.RESULT_ADD_VLAN.format(265),
            test_cisco_nexus_base.RESULT_ADD_INTERFACE.
                format('ethernet', '1\/20', 265)]),

        'delete_port2_driver_result': (
            [test_cisco_nexus_base.RESULT_DEL_INTERFACE.
            format('ethernet', '1\/20', 265),
            test_cisco_nexus_base.RESULT_DEL_VLAN.format(265)]),

        'add_port2_driver_result2': (
            [test_cisco_nexus_base.RESULT_ADD_VLAN.format(267),
            test_cisco_nexus_base.RESULT_ADD_INTERFACE.
                format('ethernet', '1\/20', 267)]),

        'delete_port2_driver_result2': (
            [test_cisco_nexus_base.RESULT_DEL_INTERFACE.
            format('ethernet', '1\/20', 267)]),

        'add_port2_driver_result3': (
            [test_cisco_nexus_base.RESULT_ADD_VLAN.format(268),
            test_cisco_nexus_base.RESULT_ADD_INTERFACE.
                format('portchannel', '2', 268),
            test_cisco_nexus_base.RESULT_ADD_VLAN.format(268),
            test_cisco_nexus_base.RESULT_ADD_INTERFACE.
                format('portchannel', '2', 268)]),

        'delete_port2_driver_result3': (
            [test_cisco_nexus_base.RESULT_DEL_INTERFACE.
            format('portchannel', '2', 268),
            test_cisco_nexus_base.RESULT_DEL_VLAN.format(268),
            test_cisco_nexus_base.RESULT_DEL_INTERFACE.
                format('portchannel', '2', 268),
            test_cisco_nexus_base.RESULT_DEL_VLAN.format(268)]),

        'add_port_channel_driver_result': (
            [test_cisco_nexus_base.RESULT_ADD_VLAN.format(268),
            test_cisco_nexus_base.RESULT_ADD_INTERFACE.
                format('portchannel', '2', 268)]),

        'delete_port_channel_driver_result': (
            [test_cisco_nexus_base.RESULT_DEL_INTERFACE.
                format('portchannel', '2', 268),
            test_cisco_nexus_base.RESULT_DEL_VLAN.format(268)]),

        'dual_add_port_driver_result': (
            [test_cisco_nexus_base.RESULT_ADD_VLAN.format(269),
            test_cisco_nexus_base.RESULT_ADD_INTERFACE.
                format('ethernet', '1\/3', 269),
            test_cisco_nexus_base.RESULT_ADD_VLAN.format(269),
            test_cisco_nexus_base.RESULT_ADD_INTERFACE.
                format('portchannel', '2', 269)]),

        'dual_delete_port_driver_result': (
            [test_cisco_nexus_base.RESULT_DEL_INTERFACE.
                format('ethernet', '1\/3', 269),
            test_cisco_nexus_base.RESULT_DEL_VLAN.format(269),
            test_cisco_nexus_base.RESULT_DEL_INTERFACE.
                format('portchannel', '2', 269)]),

        'migrate_add_host2_driver_result': (
            [test_cisco_nexus_base.RESULT_ADD_VLAN.format(267),
            test_cisco_nexus_base.RESULT_ADD_INTERFACE.
                format('ethernet', '1\/20', 267)]),

        'add_port_driver_result': (
            [test_cisco_nexus_base.RESULT_ADD_VLAN.format(267),
            test_cisco_nexus_base.RESULT_ADD_INTERFACE.
                format('ethernet', '1\/10', 267)]),

        'del_port_driver_result': (
            [test_cisco_nexus_base.RESULT_DEL_INTERFACE.
                format('ethernet', '1\/10', 267),
            test_cisco_nexus_base.RESULT_DEL_VLAN.format(267)]),

    }


class TestCiscoNexusDevice(test_cisco_nexus_base.TestCiscoNexusBase,
                           TestCiscoNexusDeviceConfig,
                           TestCiscoNexusDeviceResults):

    """Unit tests for Cisco ML2 Nexus device driver."""

    def setUp(self):
        """Sets up mock ncclient, and switch and credentials dictionaries."""

        cfg.CONF.set_override('nexus_driver', 'ncclient', 'ml2_cisco')
        cfg.CONF.set_override('never_cache_ssh_connection', False, 'ml2_cisco')
        cfg.CONF.set_override('switch_heartbeat_time', 0, 'ml2_cisco')
        super(TestCiscoNexusDevice, self).setUp()
        self.mock_ncclient.reset_mock()
        self.results = TestCiscoNexusDeviceResults()

    def test_create_delete_duplicate_ports(self):
        """Tests creation and deletion of two new virtual Ports."""

        self._basic_create_verify_port_vlan(
            'test_config1',
            self.results.get_test_results('duplicate_add_port_driver_result')
        )

        self._create_port(
            self.test_configs['test_config3'])
        # TODO(caboucha)
        # Commented out until the correct fix for
        # the following issue is resolved.
        # https://review.openstack.org/#/c/241216/
        #
        # verify first config was indeed configured
        # Original code was as follows:
        # self._verify_results(duplicate_add_port_driver_result)

        # Verify there are 2 port configs
        bindings = nexus_db_v2.get_nexusvlan_binding(
                       test_cisco_nexus_base.VLAN_ID_1,
                       test_cisco_nexus_base.NEXUS_IP_ADDRESS_1)
        self.assertEqual(2, len(bindings))

        # Clean all the ncclient mock_calls so we can evaluate
        # results of delete operations.
        self.mock_ncclient.reset_mock()

        # For results, pass empty list to verify no nexus action on
        # first port removal.
        self._basic_delete_verify_port_vlan(
            'test_config1',
            [], nbr_of_bindings=1)

        self._basic_delete_verify_port_vlan(
            'test_config3',
            self.results.get_test_results('duplicate_del_port_driver_result'))

    def test_create_delete_duplicate_port_transaction(self):
        """Tests creation and deletion same port transaction."""

        self._basic_create_verify_port_vlan(
            'test_config1',
            self.results.get_test_results('duplicate_add_port_driver_result')
        )

        self.assertEqual(
            1, len(nexus_db_v2.get_nexusport_switch_bindings(
                   test_cisco_nexus_base.NEXUS_IP_ADDRESS_1)))

        self._create_port(
            self.test_configs['test_config1'])
        self._verify_results(
            self.results.get_test_results('duplicate_add_port_driver_result')
        )

        self.assertEqual(
            1, len(nexus_db_v2.get_nexusport_switch_bindings(
                   test_cisco_nexus_base.NEXUS_IP_ADDRESS_1)))

        # Clean all the ncclient mock_calls so we can evaluate
        # results of delete operations.
        self.mock_ncclient.reset_mock()

        self._basic_delete_verify_port_vlan(
            'test_config1',
            self.results.get_test_results('duplicate_del_port_driver_result'),
            nbr_of_bindings=0)

        self._basic_delete_verify_port_vlan(
            'test_config1',
            self.results.get_test_results('duplicate_del_port_driver_result'))

    def test_create_delete_same_switch_diff_hosts_diff_vlan(self):
        """Test create/delete two Ports, same switch/diff host & vlan."""

        self._basic_create_verify_port_vlan(
            'test_config1',
            self.results.get_test_results('duplicate_add_port_driver_result'))

        self._create_port(
            self.test_configs['test_config2'])
        self._verify_results(
            self.results.get_test_results('add_port2_driver_result'))

        # Verify there are 2 port configs
        bindings = nexus_db_v2.get_nexusport_switch_bindings(
                       test_cisco_nexus_base.NEXUS_IP_ADDRESS_1)
        self.assertEqual(2, len(bindings))

        # Clean all the ncclient mock_calls so we can evaluate
        # results of delete operations.
        self.mock_ncclient.reset_mock()

        self._basic_delete_verify_port_vlan(
            'test_config2',
            self.results.get_test_results('delete_port2_driver_result'),
            nbr_of_bindings=1)

        self._basic_delete_verify_port_vlan(
            'test_config1',
            self.results.get_test_results('duplicate_del_port_driver_result'))

    def test_create_delete_same_switch_diff_hosts_same_vlan(self):
        """Test create/delete two Ports, same switch & vlan/diff host."""

        self._basic_create_verify_port_vlan(
            'test_config4',
            self.results.get_test_results('add_port_driver_result'))

        self._create_port(
            self.test_configs['test_config5'])
        self._verify_results(
            self.results.get_test_results('add_port2_driver_result2'))

        # Verify there are 2 port configs
        bindings = nexus_db_v2.get_nexusport_switch_bindings(
                       test_cisco_nexus_base.NEXUS_IP_ADDRESS_8)
        self.assertEqual(2, len(bindings))

        # Clean all the ncclient mock_calls so we can evaluate
        # results of delete operations.
        self.mock_ncclient.reset_mock()

        self._basic_delete_verify_port_vlan(
            'test_config5',
            self.results.get_test_results('delete_port2_driver_result2'),
            nbr_of_bindings=1)

        self._basic_delete_verify_port_vlan(
            'test_config4',
            self.results.get_test_results('del_port_driver_result'))

    def test_create_delete_diff_switch_same_host(self):
        """Test create/delete of two Ports, diff switch/same host."""

        self._basic_create_verify_port_vlan(
            'test_config_portchannel2',
            self.results.get_test_results('add_port2_driver_result3'))

        # Verify there are 2 port configs. One per switch.
        bindings = nexus_db_v2.get_nexusport_switch_bindings(
                       test_cisco_nexus_base.NEXUS_IP_ADDRESS_6)
        self.assertEqual(1, len(bindings))
        bindings = nexus_db_v2.get_nexusport_switch_bindings(
                       test_cisco_nexus_base.NEXUS_IP_ADDRESS_7)
        self.assertEqual(1, len(bindings))

        # Clean all the ncclient mock_calls so we can evaluate
        # results of delete operations.
        self.mock_ncclient.reset_mock()

        # For results, pass empty list to verify no nexus action on
        # first port removal.
        self._basic_delete_verify_port_vlan(
            'test_config_portchannel2',
            self.results.get_test_results('delete_port2_driver_result3'))

    def test_create_delete_portchannel(self):
        """Tests creation of a port over a portchannel."""

        self._create_delete_port(
            'test_config_portchannel',
            self.results.get_test_results('add_port_channel_driver_result'),
            self.results.get_test_results('delete_port_channel_driver_result'))

    def test_create_delete_dual(self):
        """Tests creation and deletion of dual ports for single server"""

        self._basic_create_verify_port_vlan(
            'test_config_dual',
            self.results.get_test_results('dual_add_port_driver_result'),
            nbr_of_bindings=2)

        self._basic_delete_verify_port_vlan(
            'test_config_dual',
            self.results.get_test_results('dual_delete_port_driver_result'))

    def test_create_delete_dhcp(self):
        """Tests creation and deletion of ports with device_owner of dhcp."""

        self._create_delete_port(
            'test_config_dhcp',
            self.results.get_test_results('duplicate_add_port_driver_result'),
            self.results.get_test_results('duplicate_del_port_driver_result'))

    def test_create_delete_router_ha_intf(self):
        """Tests creation and deletion of ports with device_owner
        of router_ha_interface.
        """

        self._create_delete_port(
            'test_config_router_ha_intf',
            self.results.get_test_results('duplicate_add_port_driver_result'),
            self.results.get_test_results('duplicate_del_port_driver_result'))

    def test_create_delete_router_intf(self):
        """Tests creation and deletion of ports with device_owner
        of router_interface.
        """

        self._create_delete_port(
            'test_config_router_intf',
            self.results.get_test_results('duplicate_add_port_driver_result'),
            self.results.get_test_results('duplicate_del_port_driver_result'))

    def test_create_delete_router_gateway(self):
        """Tests creation and deletion of ports with device_owner
        of router_gateway.
        """

        self._create_delete_port(
            'test_config_router_gw',
            self.results.get_test_results('duplicate_add_port_driver_result'),
            self.results.get_test_results('duplicate_del_port_driver_result'))

    def test_nexus_vm_migration(self):
        """Verify VM (live) migration.

        Simulate the following:
        Nova informs neutron of live-migration with port-update(new host).
        This should trigger two update_port_pre/postcommit() calls.

        The first one should only change the current host_id and remove the
        binding resulting in the mechanism drivers receiving:

        * PortContext.original['binding:host_id']: previous value
        * PortContext.original_top_bound_segment: previous value
        * PortContext.current['binding:host_id']: current (new) value
        * PortContext.top_bound_segment: None

        The second one binds the new host resulting in the mechanism
        drivers receiving:

        * PortContext.original['binding:host_id']: previous value
        * PortContext.original_top_bound_segment: None
        * PortContext.current['binding:host_id']: previous value
        * PortContext.top_bound_segment: new value
        """

        self._basic_create_verify_port_vlan(
            'test_config1',
            self.results.get_test_results('duplicate_add_port_driver_result'))
        binding = nexus_db_v2.get_nexusvm_bindings(
            test_cisco_nexus_base.VLAN_ID_1,
            test_cisco_nexus_base.INSTANCE_1)[0]
        self.assertEqual(
            test_cisco_nexus_base.NEXUS_PORT_1,
            binding.port_id)

        port_context = self._generate_port_context(
            self.test_configs['test_config_migrate'],
            unbind_port=True)
        port_cfg = self.test_configs['test_config1']
        port_context.set_orig_port(
            port_cfg.instance_id,
            port_cfg.host_name,
            port_cfg.device_owner,
            port_cfg.profile,
            port_cfg.vnic_type,
            test_cisco_nexus_base.NETID)

        self._cisco_mech_driver.create_port_postcommit(port_context)
        self._cisco_mech_driver.update_port_precommit(port_context)
        self._cisco_mech_driver.update_port_postcommit(port_context)

        # Verify that port entry has been deleted.
        self.assertRaises(
            exceptions.NexusPortBindingNotFound,
            nexus_db_v2.get_nexusvm_bindings,
            test_cisco_nexus_base.VLAN_ID_1,
            test_cisco_nexus_base.INSTANCE_1)

        # Clean all the ncclient mock_calls to clear exception
        # and other mock_call history.
        self.mock_ncclient.reset_mock()

        self._basic_create_verify_port_vlan(
            'test_config_migrate',
            self.results.get_test_results('migrate_add_host2_driver_result'))

        # Verify that port entry has been added using new host name.
        # Use port_id to verify that 2nd host name was used.
        binding = nexus_db_v2.get_nexusvm_bindings(
            test_cisco_nexus_base.VLAN_ID_1,
            test_cisco_nexus_base.INSTANCE_1)[0]
        self.assertEqual(
            test_cisco_nexus_base.NEXUS_PORT_2,
            binding.port_id)


class TestCiscoNexusDeviceFailure(test_cisco_nexus_base.TestCiscoNexusBase,
                           TestCiscoNexusDeviceConfig,
                           TestCiscoNexusDeviceResults):

    """Negative Unit tests for Cisco ML2 Nexus device driver."""

    def setUp(self):
        """Sets up mock ncclient, and switch and credentials dictionaries."""

        cfg.CONF.set_override('nexus_driver', 'ncclient', 'ml2_cisco')
        cfg.CONF.set_override('never_cache_ssh_connection', False, 'ml2_cisco')
        cfg.CONF.set_override('switch_heartbeat_time', 0, 'ml2_cisco')
        super(TestCiscoNexusDeviceFailure, self).setUp()
        self.mock_ncclient.reset_mock()
        self.results = TestCiscoNexusDeviceResults()

    def test_connect_failure(self):
        """Verifies exception handling during ncclient connect. """

        # Clean all the ncclient mock_calls to clear exception
        # and other mock_call history.
        self.mock_ncclient.reset_mock()

        config = {'connect.side_effect': Exception(CONNECT_ERROR)}
        self.mock_ncclient.configure_mock(**config)

        e = self.assertRaises(exceptions.NexusConnectFailed,
                              self._create_port,
                              self.test_configs[
                                  'test_config1'])
        self.assertIn(CONNECT_ERROR, six.u(str(e)))
        self.assertEqual(1, self.mock_ncclient.connect.call_count)

    def test_ncclient_fail_on_connect_other_exceptions(self):
        """Test that other errors during connect() sequences are still handled.

        If the old ncclient is installed, we expect to get a TypeError first,
        but should still handle other errors in the usual way, whether they
        appear on the first or second call to connect().
        """

        # Clear connect_call_count
        self.mock_ncclient.reset_mock()

        err_strings = ['This is TypeError',
                       'This is IOError',
                       'This is AttributeError']
        except_errors = [TypeError(err_strings[0]),
                        IOError(err_strings[1]),
                        AttributeError(err_strings[2])]
        call_count = 0
        for errors in except_errors:
            config = {'connect.side_effect': errors}
            self.mock_ncclient.configure_mock(**config)
            port_context = self._generate_port_context(
                self.test_configs['test_config1'])

            e = self.assertRaises(
                exceptions.NexusConnectFailed,
                self._cisco_mech_driver.create_port_postcommit,
                port_context)
            self.assertIn(
                "Create Failed: Port event can not "
                "be processed at this time.", six.u(str(e)))

            self._cisco_mech_driver.update_port_precommit(port_context)
            e = self.assertRaises(
                exceptions.NexusConnectFailed,
                self._cisco_mech_driver.update_port_postcommit,
                port_context)
            self.assertIn(err_strings[call_count], six.u(str(e)))

            self._cisco_mech_driver.delete_port_precommit(port_context)
            e = self.assertRaises(
                exceptions.NexusConnectFailed,
                self._cisco_mech_driver.delete_port_postcommit,
                port_context)
            self.assertIn(err_strings[call_count], six.u(str(e)))

            call_count += 1
            self.assertEqual(
                call_count * 3,
                self.mock_ncclient.connect.call_count)

    def test_get_nexus_type_failure(self):
        """Verifies exception during ncclient get inventory. """

        self._create_port_failure(
            'connect.return_value.get.side_effect',
            'show inventory',
            'test_config1',
            'Create Failed:',
            which_exc=exceptions.NexusConnectFailed)

        # Verify we attempt to connect once. get_nexus_type is a
        # special case since replay code will retry
        self.assertEqual(1, self.mock_ncclient.connect.call_count)

    def test_create_vlan_failure(self):
        """Verifies exception during edit vlan create driver. """

        self._create_port_failure(
            'connect.return_value.edit_config.side_effect',
            'vlan vlan-id-create-delete 267',
            'test_config1',
            __name__)

        # Verify we attempt to connect twice. First when first
        # create_vlan fails then _edit_config loops to attempt
        # it again and it fails again.
        self.assertEqual(2, self.mock_ncclient.connect.call_count)

    def test_delete_vlan_failure(self):
        """Verifies exception during edit vlan delete driver. """

        self._delete_port_failure(
            'connect.return_value.edit_config.side_effect',
            'vlan-id-create-delete no vlan 267',
            'test_config1',
            __name__)

    def test_create_trunk_failure(self):
        """Verifies exception during create trunk interface driver. """

        self._create_port_failure(
            'connect.return_value.edit_config.side_effect',
            'switchport trunk allowed vlan_id 267',
            'test_config1',
            __name__)

    def test_delete_trunk_failure(self):
        """Verifies exception during delete trunk interface driver. """

        self._delete_port_failure(
            'connect.return_value.edit_config.side_effect',
            'switchport trunk allowed remove vlan 267',
            'test_config1',
            __name__)

    def test_edit_fail_on_try_1(self):
        """Verifies reconnect during ncclient edit. """

        # Clean all the ncclient mock_calls so we can evaluate
        # results of delete operations.
        self.mock_ncclient.reset_mock()

        config = {'connect.return_value.edit_config.side_effect':
                  self._config_side_effects_on_count(
                      'vlan vlan-id-create-delete 267',
                      Exception(__name__), range(1))}
        self.mock_ncclient.configure_mock(**config)

        self._create_port(self.test_configs['test_config1'])

        # Verify we connected twice. Connect attempt 1 occurs on
        # any first driver call.  Then create-vlan fails first
        # time resulting close of stale handle. Driver
        # loops around to try and reopen and create-vlan should
        # then be successful on the 2nd pass.
        self.assertEqual(2, self.mock_ncclient.connect.call_count)
        self.assertEqual(1,
            self.mock_ncclient.connect.return_value.
            close_session.call_count)

    def test_nexus_host_not_configured(self):
        """Test handling of a host not found in our configuration.

        If a host is not found in the cisco configuration the driver
        should silently ignore (unknown host name is logged) and no database
        or switch configuration is performed. Exercise against all APIs.
        """

        test_func_list = [
            self._cisco_mech_driver.create_port_postcommit,
            self._cisco_mech_driver.update_port_precommit,
            self._cisco_mech_driver.update_port_postcommit,
            self._cisco_mech_driver.delete_port_precommit,
            self._cisco_mech_driver.delete_port_postcommit,
        ]

        self.mock_ncclient.reset_mock()
        port_context = self._generate_port_context(
            self.test_configs['test_config1'],
            override_host_name='no_host')

        for test_func in test_func_list:
            test_func(port_context)
            self.assertRaises(
                 exceptions.NexusPortBindingNotFound,
                 nexus_db_v2.get_nexusport_switch_bindings,
                 test_cisco_nexus_base.NEXUS_IP_ADDRESS_1)
            assert not self.mock_ncclient.connect.called

    def test_nexus_invalid_segment(self):
        """Test handling of a non VLAN segment.

        Pass a FLAT segment type into the driver. Verify that no
        exceptions are raised (non-VLAN segments are logged only) and
        that no database or switch configuration is performed.
        """

        test_func_list = [
            self._cisco_mech_driver.create_port_postcommit,
            self._cisco_mech_driver.update_port_precommit,
            self._cisco_mech_driver.update_port_postcommit,
        ]

        network_context = test_cisco_nexus_base.FakeNetworkContext(
            0, cb_constants.TYPE_FLAT)
        port_config = self.test_configs['test_config1']
        port_context = test_cisco_nexus_base.FakePortContext(
            port_config.instance_id,
            port_config.host_name,
            port_config.device_owner,
            network_context, None,
            port_config.profile,
            port_config.vnic_type
        )

        # Clear out call_count changes during initialization activity
        self.mock_ncclient.reset_mock()
        for test_func in test_func_list:
            test_func(port_context)
            self.assertRaises(
                 exceptions.NexusPortBindingNotFound,
                 nexus_db_v2.get_nexusport_switch_bindings,
                 test_cisco_nexus_base.NEXUS_IP_ADDRESS_1)
            assert not self.mock_ncclient.connect.called

    def test_nexus_missing_fields(self):
        """Test handling of a NexusMissingRequiredFields exception.

        Test the Cisco NexusMissingRequiredFields exception by using
        empty device_id value during port creation.
        """

        local_test_configs = {
            'test_config1':
                test_cisco_nexus_base.TestCiscoNexusBase.TestConfigObj(
                    test_cisco_nexus_base.NEXUS_IP_ADDRESS_1,
                    test_cisco_nexus_base.HOST_NAME_1,
                    test_cisco_nexus_base.NEXUS_PORT_1,
                    '',
                    test_cisco_nexus_base.VLAN_ID_1,
                    test_cisco_nexus_base.NO_VXLAN_ID,
                    None,
                    test_cisco_nexus_base.DEVICE_OWNER_COMPUTE,
                    {},
                    None,
                    test_cisco_nexus_base.NORMAL_VNIC),
        }

        # Clear out call_count changes during initialization activity
        self.mock_ncclient.reset_mock()
        self.assertRaises(
            exceptions.NexusMissingRequiredFields,
            self._create_port,
            local_test_configs['test_config1'])

    def test_nexus_segment_none(self):
        """Test handling of segment is None.

        Verify that None segments do not throw an exception in
        _port_action_xxx. None segments passed to the event handlers are
        logged and are not processed.
        """

        network_context = test_cisco_nexus_base.FakeNetworkContext(
            0, cb_constants.TYPE_VLAN)
        network_context._network_segments = None
        port_config = self.test_configs['test_config1']
        port_context = test_cisco_nexus_base.FakePortContext(
            port_config.instance_id,
            port_config.host_name,
            port_config.device_owner,
            network_context, None,
            port_config.profile,
            port_config.vnic_type
        )
        test_func_list = [
            self._cisco_mech_driver.update_port_precommit,
            self._cisco_mech_driver.update_port_postcommit,
        ]

        # Clear out call_count changes during initialization activity
        self.mock_ncclient.reset_mock()
        for test_func in test_func_list:
            test_func(port_context)
            self.assertRaises(
                 exceptions.NexusPortBindingNotFound,
                 nexus_db_v2.get_nexusport_switch_bindings,
                 test_cisco_nexus_base.NEXUS_IP_ADDRESS_1)
            assert not self.mock_ncclient.connect.called

    def test_nexus_ncclient_disconnect(self):
        """Test handling of closing ncclient sessions.

        When multi neutron-server processes are used verify that ncclient
        close_session method is called.
        """

        # Mock to keep track of number of close_session calls.
        ncclient_close = mock.patch.object(
            nexus_network_driver.CiscoNexusSshDriver,
            '_close_session').start()

        # Clear out call_count changes during initialization activity
        self.mock_ncclient.reset_mock()

        # Verify that ncclient is not closed by default.
        self._basic_create_verify_port_vlan(
            'test_config1',
            self.results.get_test_results('duplicate_add_port_driver_result'))
        assert not ncclient_close.called

        self._basic_delete_verify_port_vlan(
            'test_config1',
            self.results.get_test_results('duplicate_del_port_driver_result'))

        # Patch to close ncclient session.
        mock.patch.object(nexus_network_driver.CiscoNexusSshDriver,
                          '_get_close_ssh_session',
                          return_value=True).start()

        # Verify that ncclient close is called twice. Once for
        # get_nexus_type during create_port_postcommit(). Then
        # It is suppressed for successful create VLAN but called
        # after trunk interface calls.
        self._basic_create_verify_port_vlan(
            'test_config1',
            self.results.get_test_results(
                'duplicate_add_port_driver_result'))
        self.assertEqual(2, ncclient_close.call_count)

    def test_nexus_extended_vlan_range_failure(self):
        """Test that extended VLAN range config errors are ignored.

        Some versions of Nexus switch do not allow state changes for
        the extended VLAN range (1006-4094), but these errors can be
        ignored (default values are appropriate). Test that such errors
        are ignored by the Nexus plugin.
        """

        self._create_port_valid_exception(
            'connect.return_value.edit_config.side_effect',
            'vlan vlan-id-create-delete 267',
            'test_config1',
            "Can't modify state for extended")

        # No reconnect attempted...call_count will be one
        self.assertEqual(1, self.mock_ncclient.connect.call_count)

        self._create_port_valid_exception(
            'connect.return_value.edit_config.side_effect',
            'vlan vlan-id-create-delete 265',
            'test_config2',
            "Command is only allowed on VLAN")

        # No reconnect attempted...call_count will be 0 since reset_mock
        # is called in _create_port_valid_exception and caching enabled
        self.assertEqual(0, self.mock_ncclient.connect.call_count)


class TestCiscoNexusInitResults(
    test_cisco_nexus_base.TestCiscoNexusBaseResults):

    """Unit tests driver results for Cisco ML2 Nexus."""

    test_results = {
        # set 1 - switch 1.1.1.1 sets eth 1/10 & 1/20 to None
        # set 2 - switch 2.2.2.2 sets portchannel 2 to None
        # set 3 - switch 3.3.3.3 sets eth 1/20 to None
        # set 4 - switch 4.4.4.4 sets eth 1/3 & portchannel 2 to None
        # set 5 - switch 6.6.6.6 sets portchannel 2 to None
        # set 6 - switch 7.7.7.7 sets portchannel 2 to None
        # set 7 - switch 8.8.8.8 sets eth 1/10 & 1/20 to None
        'duplicate_init_port_driver_result1': (
            [test_cisco_nexus_base.RESULT_INTERFACE.
                format('ethernet', '1\/10', 'None'),
            test_cisco_nexus_base.RESULT_INTERFACE.
                format('portchannel', '2', 'None'),
            test_cisco_nexus_base.RESULT_INTERFACE.
                format('ethernet', '1\/20', 'None'),
            test_cisco_nexus_base.RESULT_INTERFACE.
                format('ethernet', '1\/3', 'None'),
            test_cisco_nexus_base.RESULT_INTERFACE.
                format('portchannel', '2', 'None'),
            test_cisco_nexus_base.RESULT_INTERFACE.
                format('portchannel', '2', 'None'),
            test_cisco_nexus_base.RESULT_INTERFACE.
                format('ethernet', '1\/10', 'None')]),

        # Only one entry to match for last 3 so make None
        # so count matches in _verify_results
        'duplicate_init_port_driver_result2': (
            [test_cisco_nexus_base.RESULT_INTERFACE.
                format('ethernet', '1\/20', 'None'),
            None,
            None,
            test_cisco_nexus_base.RESULT_INTERFACE.
                format('portchannel', '2', 'None'),
            None,
            None,
            test_cisco_nexus_base.RESULT_INTERFACE.
                format('ethernet', '1\/20', 'None')]),
    }


class TestCiscoNexusDeviceInit(test_cisco_nexus_base.TestCiscoNexusBase,
                               TestCiscoNexusDeviceConfig):
    """Verifies interface vlan allowed none is set when missing."""

    def mock_init(self):
        # Prevent default which returns
        # 'switchport trunk allowed vlan none'
        # in get interface calls so Nexus driver
        # initialization will send it to Nexus device.
        data_xml = {'connect.return_value.get.return_value.data_xml':
                    ''}
        self.mock_ncclient.configure_mock(**data_xml)

    def setUp(self):
        """Sets up mock ncclient, and switch and credentials dictionaries."""

        cfg.CONF.set_override('nexus_driver', 'ncclient', 'ml2_cisco')
        cfg.CONF.set_override('never_cache_ssh_connection', False, 'ml2_cisco')
        cfg.CONF.set_override('switch_heartbeat_time', 0, 'ml2_cisco')
        super(TestCiscoNexusDeviceInit, self).setUp()
        self.results = TestCiscoNexusInitResults()

    def test_verify_initialization(self):

        self._verify_results(
            self.results.get_test_results(
                'duplicate_init_port_driver_result1'))
        self._verify_results(
            self.results.get_test_results(
                'duplicate_init_port_driver_result2'))


class TestCiscoNexusBaremetalResults(
    test_cisco_nexus_base.TestCiscoNexusBaseResults):

    """Unit tests driver results for Cisco ML2 Nexus."""

    test_results = {
        'add_port_ethernet_driver_result': (
            [test_cisco_nexus_base.RESULT_ADD_VLAN.format(267),
            (test_cisco_nexus_base.RESULT_ADD_NATIVE_INTERFACE.
                format('ethernet', '1\/10', 267) +
            '[\x00-\x7f]+' +
            test_cisco_nexus_base.RESULT_ADD_INTERFACE.
                format('ethernet', '1\/10', 267))]),

        'delete_port_ethernet_driver_result': (
            [(test_cisco_nexus_base.RESULT_DEL_NATIVE_INTERFACE.
                format('ethernet', '1\/10') +
            '[\x00-\x7f]+' +
             test_cisco_nexus_base.RESULT_DEL_INTERFACE.
                format('ethernet', '1\/10', 267)),
            test_cisco_nexus_base.RESULT_DEL_VLAN.format(267)]),

        'add_vm_port_ethernet_driver_result': (
            [test_cisco_nexus_base.RESULT_ADD_VLAN.format(265),
            test_cisco_nexus_base.RESULT_ADD_INTERFACE.
                format('ethernet', '1\/10', 265)]),

        'delete_vm_port_ethernet_driver_result': (
            [test_cisco_nexus_base.RESULT_DEL_INTERFACE.
                format('ethernet', '1\/10', 265),
            test_cisco_nexus_base.RESULT_DEL_VLAN.format(265)]),

        'add_port_ethernet_native_driver_result': (
            [test_cisco_nexus_base.RESULT_ADD_VLAN.format(265),
            (test_cisco_nexus_base.RESULT_ADD_NATIVE_INTERFACE.
            format('ethernet', '1\/10', 265) +
            '[\x00-\x7f]+' +
            test_cisco_nexus_base.RESULT_ADD_INTERFACE.
                format('ethernet', '1\/10', 265))]),

        'delete_port_ethernet_native_driver_result': (
            [(test_cisco_nexus_base.RESULT_DEL_NATIVE_INTERFACE.
                format('ethernet', '1\/10') +
            '[\x00-\x7f]+' +
            test_cisco_nexus_base.RESULT_DEL_INTERFACE.
                format('ethernet', '1\/10', 265)),
            test_cisco_nexus_base.RESULT_DEL_VLAN.format(265)]),

        'driver_result_unique_vPC_add1': (
            [test_cisco_nexus_base.RESULT_ADD_VLAN.format(267),
            (test_cisco_nexus_base.RESULT_ADD_NATIVE_INTERFACE.
            format('port-channel', '469', 267) +
            '[\x00-\x7f]+' +
            test_cisco_nexus_base.RESULT_ADD_INTERFACE.
                format('port-channel', '469', 267)),
            test_cisco_nexus_base.RESULT_ADD_VLAN.format(267),
            (test_cisco_nexus_base.RESULT_ADD_NATIVE_INTERFACE.
            format('port-channel', '469', 267) +
            '[\x00-\x7f]+' +
            test_cisco_nexus_base.RESULT_ADD_INTERFACE.
                format('port-channel', '469', 267))]),

        'driver_result_unique_vPC_del1': (
            [(test_cisco_nexus_base.RESULT_DEL_NATIVE_INTERFACE.
                format('port-channel', '469') +
            '[\x00-\x7f]+' +
            test_cisco_nexus_base.RESULT_DEL_INTERFACE.
                format('port-channel', '469', 267)),
            test_cisco_nexus_base.RESULT_DEL_VLAN.format(267),
            (test_cisco_nexus_base.RESULT_DEL_NATIVE_INTERFACE.
                format('port-channel', '469') +
            '[\x00-\x7f]+' +
            test_cisco_nexus_base.RESULT_DEL_INTERFACE.
                format('port-channel', '469', 267)),
            test_cisco_nexus_base.RESULT_DEL_VLAN.format(267)]),

        'driver_result_unique_vPC_add1_vm': (
            [test_cisco_nexus_base.RESULT_ADD_VLAN.format(265),
            test_cisco_nexus_base.RESULT_ADD_INTERFACE.
                format('port-channel', '469', 265),
            test_cisco_nexus_base.RESULT_ADD_VLAN.format(265),
            test_cisco_nexus_base.RESULT_ADD_INTERFACE.
                format('port-channel', '469', 265)]),

        'driver_result_unique_vPC_del1_vm': (
            [test_cisco_nexus_base.RESULT_DEL_INTERFACE.
                format('port-channel', '469', 265),
            test_cisco_nexus_base.RESULT_DEL_VLAN.format(265),
            test_cisco_nexus_base.RESULT_DEL_INTERFACE.
                format('port-channel', '469', 265),
            test_cisco_nexus_base.RESULT_DEL_VLAN.format(265)]),

    }


class TestCiscoNexusBaremetalDevice(test_cisco_nexus_base.TestCiscoNexusBase):

    """Unit tests for Cisco ML2 Nexus baremetal device driver."""

    baremetal_profile = {
        "local_link_information": [
            {
                "port_id": test_cisco_nexus_base.NEXUS_BAREMETAL_PORT_1,
                "switch_info": {
                    "switch_ip": test_cisco_nexus_base.NEXUS_IP_ADDRESS_1,
                },
            },
        ]
    }

    baremetal_profile_vPC = {
        "local_link_information": [
            {
                "port_id": test_cisco_nexus_base.NEXUS_BAREMETAL_PORT_1,
                "switch_info": {
                    "switch_ip": test_cisco_nexus_base.NEXUS_IP_ADDRESS_1,
                },
            },
            {
                "port_id": test_cisco_nexus_base.NEXUS_BAREMETAL_PORT_2,
                "switch_info": {
                    "switch_ip": test_cisco_nexus_base.NEXUS_IP_ADDRESS_2,
                },
            },
        ]
    }

    baremetal_profile_is_native = {
        "local_link_information": [
            {
                "port_id": test_cisco_nexus_base.NEXUS_BAREMETAL_PORT_1,
                "switch_info": {
                    "switch_ip": test_cisco_nexus_base.NEXUS_IP_ADDRESS_1,
                },
            },
        ]
    }

    # The IP Address and Nexus Port information is duplicated in case
    # of baremetal.  The core code uses content of baremetal_profile
    # While test code continues to use values in test_config
    # for verification. This keeps test code simpler.
    test_configs = {
        'test_config1':
            test_cisco_nexus_base.TestCiscoNexusBase.TestConfigObj(
                test_cisco_nexus_base.NEXUS_IP_ADDRESS_1,
                test_cisco_nexus_base.HOST_NAME_UNUSED,
                test_cisco_nexus_base.NEXUS_BAREMETAL_PORT_1,
                test_cisco_nexus_base.INSTANCE_1,
                test_cisco_nexus_base.VLAN_ID_1,
                test_cisco_nexus_base.NO_VXLAN_ID,
                None,
                test_cisco_nexus_base.DEVICE_OWNER_BAREMETAL,
                baremetal_profile,
                test_cisco_nexus_base.HOST_NAME_Baremetal + '1',
                test_cisco_nexus_base.BAREMETAL_VNIC),
        'test_config_vm':
            test_cisco_nexus_base.TestCiscoNexusBase.TestConfigObj(
                test_cisco_nexus_base.NEXUS_IP_ADDRESS_1,
                test_cisco_nexus_base.HOST_NAME_Baremetal + '1',
                test_cisco_nexus_base.NEXUS_BAREMETAL_PORT_1,
                test_cisco_nexus_base.INSTANCE_2,
                test_cisco_nexus_base.VLAN_ID_2,
                test_cisco_nexus_base.NO_VXLAN_ID,
                None,
                test_cisco_nexus_base.DEVICE_OWNER_COMPUTE,
                {},
                None,
                test_cisco_nexus_base.NORMAL_VNIC),
        'test_config_vPC':
            test_cisco_nexus_base.TestCiscoNexusBase.TestConfigObj(
                None,
                test_cisco_nexus_base.HOST_NAME_UNUSED,
                test_cisco_nexus_base.NEXUS_BAREMETAL_PORT_1,
                test_cisco_nexus_base.INSTANCE_1,
                test_cisco_nexus_base.VLAN_ID_1,
                test_cisco_nexus_base.NO_VXLAN_ID,
                None,
                test_cisco_nexus_base.DEVICE_OWNER_BAREMETAL,
                baremetal_profile_vPC,
                test_cisco_nexus_base.HOST_NAME_Baremetal + '1',
                test_cisco_nexus_base.BAREMETAL_VNIC),
        'test_config_native':
            test_cisco_nexus_base.TestCiscoNexusBase.TestConfigObj(
                test_cisco_nexus_base.NEXUS_IP_ADDRESS_1,
                test_cisco_nexus_base.HOST_NAME_UNUSED,
                test_cisco_nexus_base.NEXUS_BAREMETAL_PORT_1,
                test_cisco_nexus_base.INSTANCE_1,
                test_cisco_nexus_base.VLAN_ID_2,
                test_cisco_nexus_base.NO_VXLAN_ID,
                None,
                test_cisco_nexus_base.DEVICE_OWNER_BAREMETAL,
                baremetal_profile_is_native,
                None,
                test_cisco_nexus_base.BAREMETAL_VNIC),
    }

    def setUp(self):
        """Sets up mock ncclient, and switch and credentials dictionaries."""

        cfg.CONF.set_override('nexus_driver', 'ncclient', 'ml2_cisco')
        cfg.CONF.set_override('never_cache_ssh_connection', False, 'ml2_cisco')
        cfg.CONF.set_override('switch_heartbeat_time', 0, 'ml2_cisco')
        super(TestCiscoNexusBaremetalDevice, self).setUp()
        self.results = TestCiscoNexusBaremetalResults()

    def _init_port_channel(self, which=1):
        # this is to prevent interface initialization from occurring
        # which adds unnecessary noise to the results.
        data_xml = {'connect.return_value.get.return_value.data_xml':
                    'switchport trunk allowed vlan none\n'
                    'channel-group 469 mode active'}
        self.mock_ncclient.configure_mock(**data_xml)

    def test_create_delete_basic_bm_ethernet_port_and_vm(self):
        """Basic creation and deletion test of 1 ethernet port."""

        self._basic_create_verify_port_vlan(
            'test_config1',
            self.results.get_test_results(
                'add_port_ethernet_driver_result'))

        # Clean all driver mock_calls so we can evaluate next
        # set of results.
        self.mock_ncclient.reset_mock()

        self._basic_create_verify_port_vlan(
            'test_config_vm',
            self.results.get_test_results(
                'add_vm_port_ethernet_driver_result'),
            nbr_of_bindings=2)

        self._basic_delete_verify_port_vlan(
            'test_config_vm',
            self.results.get_test_results(
                'delete_vm_port_ethernet_driver_result'),
            nbr_of_bindings=1)

        self._basic_delete_verify_port_vlan(
            'test_config1',
            self.results.get_test_results(
                'delete_port_ethernet_driver_result'))

    def test_create_delete_basic_port_channel(self):
        """Basic creation and deletion test of 1 port-channel."""

        self._init_port_channel(3)
        self._basic_create_verify_port_vlan(
            'test_config_vPC',
            self.results.get_test_results(
                'driver_result_unique_vPC_add1'),
            nbr_of_bindings=2)

        # Clean all the ncclient mock_calls so we can evaluate
        # results of delete operations.
        self.mock_ncclient.reset_mock()

        self._basic_delete_verify_port_vlan(
            'test_config_vPC',
            self.results.get_test_results(
                'driver_result_unique_vPC_del1'))

    def test_create_delete_learn_vpc_and_vm(self):
        """Basic creation and deletion test of 2 learn port-channel and vm."""

        switch_list = ['1.1.1.1', '2.2.2.2']

        self._init_port_channel(3)

        for switch_ip in switch_list:
            cfg.CONF.set_override(
                const.VPCPOOL, ('451-475'),
                cfg.CONF.ml2_cisco.nexus_switches.get(switch_ip)._group)
        self._cisco_mech_driver._initialize_vpc_alloc_pools()

        self._basic_create_verify_port_vlan(
            'test_config_vPC',
            self.results.get_test_results(
                'driver_result_unique_vPC_add1'),
            nbr_of_bindings=2)

        # Clean all the ncclient mock_calls so we can evaluate
        # results of delete operations.
        self.mock_ncclient.reset_mock()

        self._basic_create_verify_port_vlan(
            'test_config_vm',
            self.results.get_test_results(
                'driver_result_unique_vPC_add1_vm'),
            nbr_of_bindings=4)

        self._basic_delete_verify_port_vlan(
            'test_config_vm',
            self.results.get_test_results(
                'driver_result_unique_vPC_del1_vm'),
            nbr_of_bindings=2)

        self._basic_delete_verify_port_vlan(
            'test_config_vPC',
            self.results.get_test_results(
                'driver_result_unique_vPC_del1'))

        for switch_ip in switch_list:
            self.assertEqual(
                25, len(nexus_db_v2.get_free_switch_vpc_allocs(switch_ip)))

    def test_create_delete_basic_eth_port_is_native(self):
        """Basic creation and deletion test of 1 ethernet port."""

        self._basic_create_verify_port_vlan(
            'test_config_native',
            self.results.get_test_results(
                'add_port_ethernet_native_driver_result'))

        # Clean all the ncclient mock_calls so we can evaluate
        # results of delete operations.
        self.mock_ncclient.reset_mock()

        self._basic_delete_verify_port_vlan(
            'test_config_native',
            self.results.get_test_results(
                'delete_port_ethernet_native_driver_result'))

    def test_create_delete_switch_ip_not_defined(self):
        """Create/delete of 1 ethernet port switchinfo is string."""

        baremetal_profile_no_switch_ip = {
            "local_link_information": [
                # This IP is configured at init time
                {
                    "port_id": test_cisco_nexus_base.NEXUS_BAREMETAL_PORT_1,
                    "switch_info": {
                        "switch_ip": "1.1.1.1",
                    },
                },
                # This IP not configured at init time
                {
                    "port_id": test_cisco_nexus_base.NEXUS_BAREMETAL_PORT_1,
                    "switch_info": {
                        "switch_ip": "6.6.6.6",
                    },
                },
            ]
        }

        local_test_configs = {
            'test_config1':
                test_cisco_nexus_base.TestCiscoNexusBase.TestConfigObj(
                    test_cisco_nexus_base.NEXUS_IP_ADDRESS_1,
                    test_cisco_nexus_base.HOST_NAME_UNUSED,
                    test_cisco_nexus_base.NEXUS_BAREMETAL_PORT_1,
                    test_cisco_nexus_base.INSTANCE_1,
                    test_cisco_nexus_base.VLAN_ID_1,
                    test_cisco_nexus_base.NO_VXLAN_ID,
                    None,
                    test_cisco_nexus_base.DEVICE_OWNER_BAREMETAL,
                    baremetal_profile_no_switch_ip,
                    test_cisco_nexus_base.HOST_NAME_Baremetal + '3',
                    test_cisco_nexus_base.BAREMETAL_VNIC),
        }

        self._basic_create_verify_port_vlan(
            '',
            self.results.get_test_results(
                'add_port_ethernet_driver_result'), 1,
            other_test=local_test_configs['test_config1'])

        # Clean all the ncclient mock_calls so we can evaluate
        # results of delete operations.
        self.mock_ncclient.reset_mock()

        self._basic_delete_verify_port_vlan(
            '',
            self.results.get_test_results(
                'delete_port_ethernet_driver_result'),
            nbr_of_bindings=0,
            other_test=local_test_configs['test_config1'])


class TestCiscoNexusNonCacheSshDevice(
    test_cisco_nexus_base.TestCiscoNexusBase):

    """Unit tests for Cisco ML2 Nexus device driver in non-cache ssh mode."""

    # Testing new default of True for config var 'never_cache_ssh_connection'

    test_configs = {
        'test_config1':
            test_cisco_nexus_base.TestCiscoNexusBase.TestConfigObj(
                test_cisco_nexus_base.NEXUS_IP_ADDRESS_1,
                test_cisco_nexus_base.HOST_NAME_1,
                test_cisco_nexus_base.NEXUS_PORT_1,
                test_cisco_nexus_base.INSTANCE_1,
                test_cisco_nexus_base.VLAN_ID_1,
                test_cisco_nexus_base.NO_VXLAN_ID,
                None,
                test_cisco_nexus_base.DEVICE_OWNER_COMPUTE,
                {},
                None,
                test_cisco_nexus_base.NORMAL_VNIC),
    }

    simple_add_port_ethernet_driver_result = (
        [test_cisco_nexus_base.RESULT_ADD_VLAN.format(267),
        test_cisco_nexus_base.RESULT_ADD_INTERFACE.
            format('ethernet', '1\/10', 267)])

    simple_delete_port_ethernet_driver_result = (
        [test_cisco_nexus_base.RESULT_DEL_INTERFACE.
            format('ethernet', '1\/10', 267),
        test_cisco_nexus_base.RESULT_DEL_VLAN.format(267)])

    def setUp(self):
        cfg.CONF.set_override('nexus_driver', 'ncclient', 'ml2_cisco')
        super(TestCiscoNexusNonCacheSshDevice, self).setUp()
        cfg.CONF.set_override('switch_heartbeat_time', 0, 'ml2_cisco')
        self.mock_ncclient.reset_mock()

    def test_create_delete_basic(self):
        """Basic creation and deletion test of 1 ethernet port."""

        # Clean all the ncclient mock_calls so we can evaluate
        # results of add operation.
        self.mock_ncclient.reset_mock()

        # Call _create_port directly without verification
        # We know at this point that this works.
        self._create_port(self.test_configs['test_config1'])

        # The objective is to verify call count when caching disabled
        self.assertEqual(2, self.mock_ncclient.connect.call_count)
        self.assertEqual(2,
            self.mock_ncclient.connect.return_value.
            close_session.call_count)

        # Clean all the ncclient mock_calls so we can evaluate
        # results of delete operations.
        self.mock_ncclient.reset_mock()

        # nbr_of_bindings includes reserved port binding
        self._delete_port(self.test_configs['test_config1'])

        self.assertEqual(2, self.mock_ncclient.connect.call_count)
        self.assertEqual(2,
            self.mock_ncclient.connect.return_value.
            close_session.call_count)

    def test_edit_fail_on_try_1(self):
        """Verifies reconnect during ncclient edit. """

        # Clean all the ncclient mock_calls so we can evaluate
        # results of delete operations.
        self.mock_ncclient.reset_mock()

        config = {'connect.return_value.edit_config.side_effect':
                  self._config_side_effects_on_count(
                      'vlan vlan-id-create-delete 267',
                      Exception(__name__), range(1))}
        self.mock_ncclient.configure_mock(**config)

        self._create_port(self.test_configs['test_config1'])

        # With ssh handle not patched, there will be 3 connects
        # and 3 closes.
        # 1) Full connect during create_port get nexus type call
        #    and close after this call.
        # 2) Full connect during update_port on first failed
        #    create_vlan, then close on error. Driver then
        #    loops back and performs a full reconnect on
        #    successful send of create_vlan.
        #    The close operation is skipped following this.
        # 3) When interface configuration is sent, a close
        #    is then perform to complete this transaction set.
        self.assertEqual(3, self.mock_ncclient.connect.call_count)
        self.assertEqual(3,
            self.mock_ncclient.connect.return_value.
            close_session.call_count)



# Copyright 2016 Cisco Systems, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

#
# Openstack driver for cisco ASA firewall
#

import base64
import netaddr
from networking_cisco.apps.saf.common import dfa_logger as logging
from oslo_serialization import jsonutils
try:
    import urllib2
except ImportError:
    import urllib.request as urllib2

LOG = logging.getLogger(__name__)


class Asa5585(object):
    """ASA 5585 Driver. """

    def __init__(self, mgmt_ip, username, password):
        self.server = "https://" + mgmt_ip
        self.username = username
        self.password = password
        self.tenant_rule = dict()
        self.rule_tbl = dict()

    def rest_send_cli(self, data):
        headers = {'Content-Type': 'application/json'}
        api_path = "/api/cli"    # param
        url = self.server + api_path

        # Use request library instead of urllib2. Also for status codes,
        # use the literals provided by request library, instead of hardcoding.
        # TODO(padkrish)
        req = urllib2.Request(url, jsonutils.dumps(data), headers)
        byte_str = ('%s:%s' % ('user', 'user')).encode()
        base64string = base64.encodestring(byte_str).decode().replace('\n', '')

        req.add_header("Authorization", "Basic %s" % base64string)
        f = None
        status_code = 400
        try:
            f = urllib2.urlopen(req)
            status_code = f.getcode()
            LOG.info("Status code is %d", status_code)

        except (urllib2.HTTPError, netaddr.err):
            LOG.error("Error received from server. HTTP status code "
                      "is %d", netaddr.err.code)
            try:
                json_error = jsonutils.loads(netaddr.err.read())
                if json_error:
                    LOG.error("Error in Json Loads",
                              jsonutils.dumps(json_error, sort_keys=True,
                                              indent=4,
                                              separators=(',', ': ')))
            except ValueError:
                pass
        finally:
            if f:
                f.close()
        return (status_code in range(200, 300))

    def setup(self, **kwargs):
        """setup ASA context for an edge tenant pair. """
        params = kwargs.get('params')
        LOG.info("asa_setup: tenant %(tenant)s %(in_vlan)d %(out_vlan)d"
                 " %(in_ip)s %(in_mask)s %(out_ip)s %(out_mask)s",
                 {'tenant': params.get('tenant_name'),
                  'in_vlan': params.get('in_vlan'),
                  'out_vlan': params.get('out_vlan'),
                  'in_ip': params.get('in_ip'),
                  'in_mask': params.get('in_mask'),
                  'out_ip': params.get('out_ip'),
                  'out_mask': params.get('out_mask')})
        inside_vlan = str(params.get('in_vlan'))
        outside_vlan = str(params.get('out_vlan'))
        context = params.get('tenant_name')
        cmds = ["conf t", "changeto system"]
        inside_int = params.get('intf_in') + '.' + inside_vlan
        cmds.append("int " + inside_int)
        cmds.append("vlan " + inside_vlan)
        outside_int = params.get('intf_out') + '.' + outside_vlan
        cmds.append("int " + outside_int)
        cmds.append("vlan " + outside_vlan)
        cmds.append("context " + context)
        cmds.append("allocate-interface " + inside_int)
        cmds.append("allocate-interface " + outside_int)
        cmds.append("config-url disk0:/" + context + ".cfg")
        cmds.append("write memory")
        cmds.append("changeto context " + context)
        cmds.append("int " + inside_int)
        cmds.append("nameif Inside")
        cmds.append("security-level 100")
        cmds.append(
            "ip address " + params.get('in_ip') + " " + params.get('in_mask'))
        cmds.append("int " + outside_int)
        cmds.append("nameif Outside")
        cmds.append("security-level 0")
        cmds.append("ip address " + params.get('out_ip') + " " +
                    params.get('out_mask'))

        cmds.append("router ospf 1")
        cmds.append("network " + params.get('in_ip') + " " +
                    params.get('in_mask') + " area 0")
        cmds.append("network " + params.get('out_ip') + " " +
                    params.get('out_mask') + " area 0")
        cmds.append("area 0")
        cmds.append("route Outside 0.0.0.0 0.0.0.0 " + params.get('out_gw') +
                    " 1")
        cmds.append("route Outside 0.0.0.0 0.0.0.0 " +
                    params.get('out_sec_gw') + " 1")
        cmds.append("end")
        cmds.append("write memory")

        if context not in self.tenant_rule:
            self.tenant_rule[context] = dict()
            self.tenant_rule[context]['rule_lst'] = []

        data = {"commands": cmds}
        return self.rest_send_cli(data)

    def cleanup(self, **kwargs):
        """cleanup ASA context for an edge tenant pair. """
        params = kwargs.get('params')
        LOG.info("asa_cleanup: tenant %(tenant)s %(in_vlan)d %(out_vlan)d"
                 " %(in_ip)s %(in_mask)s %(out_ip)s %(out_mask)s",
                 {'tenant': params.get('tenant_name'),
                  'in_vlan': params.get('in_vlan'),
                  'out_vlan': params.get('out_vlan'),
                  'in_ip': params.get('in_ip'),
                  'in_mask': params.get('in_mask'),
                  'out_ip': params.get('out_ip'),
                  'out_mask': params.get('out_mask')})
        inside_vlan = str(params.get('in_vlan'))
        outside_vlan = str(params.get('out_vlan'))
        context = params.get('tenant_name')
        cmds = ["conf t", "changeto system"]
        cmds.append("no context " + context + " noconfirm")
        inside_int = params.get('intf_in') + '.' + inside_vlan
        outside_int = params.get('intf_out') + '.' + outside_vlan
        cmds.append("no interface " + inside_int)
        cmds.append("no interface " + outside_int)
        cmds.append("write memory")
        cmds.append("del /noconfirm disk0:/" + context + ".cfg")

        if context in self.tenant_rule:
            for rule in self.tenant_rule[context].get('rule_lst'):
                del self.rule_tbl[rule]
            del self.tenant_rule[context]
        data = {"commands": cmds}
        return self.rest_send_cli(data)

    def get_quota(self):
        cmds = ["conf t", "changeto system"]
        cmds.append("show ver | grep Contexts")
        data = {"commands": cmds}
        headers = {'Content-Type': 'application/json'}
        api_path = "/api/cli"    # param
        url = self.server + api_path

        req = urllib2.Request(url, jsonutils.dumps(data), headers)
        byte_str = ('%s:%s' % ('user', 'user')).encode()
        base64string = base64.encodestring(byte_str).decode().replace('\n', '')
        req.add_header("Authorization", "Basic %s" % base64string)
        max_ctx_count = 0
        f = None
        try:
            f = urllib2.urlopen(req)
            status_code = f.getcode()
            LOG.info("Status code is %d", status_code)
            if status_code in range(200, 300):
                resp = jsonutils.loads(f.read())
                try:
                    max_ctx_count = int(resp.get('response')[-1].split()[3])
                except ValueError:
                    max_ctx_count = 0
                LOG.info("Max Context Count is %d", max_ctx_count)

        except (urllib2.HTTPError, netaddr.err):
            LOG.info("Error received from server. HTTP status code is %d",
                     netaddr.err.code)
            try:
                json_error = jsonutils.loads(netaddr.err.read())
                if json_error:
                    LOG.info("Error in Json loads",
                             jsonutils.dumps(json_error, sort_keys=True,
                                             indent=4,
                                             separators=(',', ': ')))
            except ValueError:
                pass
        finally:
            if f:
                f.close()
        return max_ctx_count

    def get_ip_address(self, ip_address):
        """Decode the IP address. """
        if ip_address is None:
            ip_address = '0.0.0.0/0'
        return netaddr.IPNetwork(ip_address)

    def build_acl_ip(self, network_obj):
        "Build the acl for IP address. "

        if str(network_obj) == '0.0.0.0/0':
            acl = "any "
        else:
            acl = "%(ip)s %(mask)s " % {'ip': network_obj.network,
                                        'mask': network_obj.netmask}
        return acl

    def build_acl_port(self, port, enabled=True):
        "Build the acl for L4 Ports. "
        if port is not None:
            if ':' in port:
                range = port.replace(':', ' ')
                acl = "range %(range)s " % {'range': range}
            else:
                acl = "eq %(port)s " % {'port': port}
                if not enabled:
                    acl += "inactive"
            return acl

    def build_acl(self, tenant_name, rule):
        """Build the ACL. """
        # TODO(padkrish) actions that is not deny or allow, throw error
        if rule['action'] == 'allow':
            action = 'permit'
        else:
            action = 'deny'
        acl_str = "access-list %(tenant)s extended %(action)s %(prot)s "
        acl = acl_str % {'tenant': tenant_name, 'action': action,
                         'prot': rule.get('protocol')}
        src_ip = self.get_ip_address(rule.get('source_ip_address'))
        ip_acl = self.build_acl_ip(src_ip)
        acl += ip_acl

        acl += self.build_acl_port(rule.get('source_port'))

        dst_ip = self.get_ip_address(rule.get('destination_ip_address'))
        ip_acl = self.build_acl_ip(dst_ip)
        acl += ip_acl

        acl += self.build_acl_port(rule.get('destination_port'),
                                   enabled=rule.get('enabled'))
        return acl

    def apply_policy(self, policy):
        """Apply a firewall policy. """
        tenant_name = policy['tenant_name']
        fw_id = policy['fw_id']
        fw_name = policy['fw_name']
        LOG.info("asa_apply_policy: tenant=%(tenant)s fw_id=%(fw_id)s "
                 "fw_name=%(fw_name)s",
                 {'tenant': tenant_name, 'fw_id': fw_id, 'fw_name': fw_name})
        cmds = ["conf t", "changeto context " + tenant_name]

        for rule_id, rule in policy['rules'].items():
            acl = self.build_acl(tenant_name, rule)

            LOG.info("rule[%(rule_id)s]: name=%(name)s enabled=%(enabled)s"
                     " protocol=%(protocol)s dport=%(dport)s "
                     "sport=%(sport)s dip=%(dport)s "
                     "sip=%(sip)s action=%(dip)s",
                     {'rule_id': rule_id, 'name': rule.get('name'),
                      'enabled': rule.get('enabled'),
                      'protocol': rule.get('protocol'),
                      'dport': rule.get('dst_port'),
                      'sport': rule.get('src_port'),
                      'dip': rule.get('destination_ip_address'),
                      'sip': rule.get('source_ip_address'),
                      'action': rule.get('action')})

            # remove the old ace for this rule
            if rule_id in self.rule_tbl:
                cmds.append('no ' + self.rule_tbl[rule_id])

            self.rule_tbl[rule_id] = acl
            if tenant_name in self.tenant_rule:
                if rule_id not in self.tenant_rule[tenant_name]['rule_lst']:
                    self.tenant_rule[tenant_name]['rule_lst'].append(rule_id)
            cmds.append(acl)
        cmds.append("access-group " + tenant_name + " global")
        cmds.append("write memory")

        LOG.info("cmds sent is %s", cmds)
        data = {"commands": cmds}
        return self.rest_send_cli(data)



# Copyright 2016 Cisco Systems, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

# A tiny and simple Cisco IOS XE running config simulator.
# The intended use is to allow a developer to observe how the running config
# of an IOS XE device evolves as CLI commands are issued.
#
# Simple implies here that no CLI syntax or semantical checks are made so it
# is entirely up to the command issuer to ensure the correctness of the
# commands and their arguments.
#
# Bob Melander (bob.melander@gmail.com)

import re
import six

from oslo_utils import timeutils


class CiscoIOSXESimulator(object):

    # set of commands to be logged only
    log_only_commands = set()
    # set of commands bound to immediately preceding line
    parent_bound_commands = {'exit-address-family'}
    exclamation = {'vrf definition', ' exit-address-family'}

    def __init__(self, path, host_ip, netmask, port, username, password,
                 device_params, mgmt_interface, timeout):
        self.host_ip = host_ip
        self.netmask = netmask
        self.port = port
        self.username = username
        self.password = password
        self.device_params = device_params
        self.mgmt_interface = mgmt_interface
        self.timeout = timeout
        self.last_update = timeutils.utcnow()
        self.rc = self._get_dict()
        self._set_default_config()

    def get_config(self):
        change_date = self.last_update.strftime('%a %b %d %Y')
        change_time = self.last_update.strftime('%H:%M:%S')
        intro_lines = ("! Last configuration change at " + change_time + " " +
                       "UTC " + change_date + " by " + self.username + "\n!\n")
        intro_lines += ("hostname ASR-1002X-" +
                        self.host_ip.replace('.', '_') + "\n!\n")
        intro_lines += ("boot-start-marker\nboot system flash "
                        "bootflash:/asr1002x-simulated.03.16.00.S-ext."
                        "SPA.bin\nboot-end-marker\n!\n")
        rc_data = {'rc_str': intro_lines}
        cmds = sorted(self.rc.keys())
        if 'vrf' in cmds:
            cmds.remove('vrf')
            cmds.insert(0, 'vrf')
        #for cmd, args in sorted(six.iteritems(self.rc)):
        for cmd in cmds:
            args = self.rc[cmd]
            line = cmd
            self._build_line(rc_data, args, line, 0)
        print(rc_data['rc_str'])
        return rc_data['rc_str']

    def edit_config(self, snippet):
        command_lines = self._get_command_lines(snippet)
        if not command_lines:
            return
        self._process_next_level(self.rc, self.rc, command_lines, None, True)
        self.last_update = timeutils.utcnow()
        return True

    def _set_default_config(self):
        mgmt_gw_ip = '.'.join((self.host_ip.split('.')[:-1]) + ['1'])
        command_chunks = [
            ["vrf definition Mgmt-intf",
             "address-family ipv4",
             "exit-address-family",
             "address-family ipv6",
             "exit-address-family"],
            ["interface " + self.mgmt_interface,
             "vrf forwarding Mgmt-intf",
             "ip address " + self.host_ip + " " + self.netmask,
             "negotiation auto"],
            ["ip tftp source - interface " + self.mgmt_interface],
            ["ip route vrf Mgmt - intf 0.0.0.0 0.0.0.0 " + mgmt_gw_ip],
            ["ip ssh source - interface " + self.mgmt_interface],
            ["ip ssh version 2"]
        ]
        for commands in command_chunks:
            self._process_next_level(self.rc, self.rc, commands, None, True)

    def _build_line(self, rc_data, current, baseline, level):
        for string, the_rest in sorted(six.iteritems(current)):
            if string == 'EOL':
                continue
            #line = baseline
            line = " " if baseline == "" and level >= 1 else baseline
            #line += ' ' + string if line != "" else string
            line += ' ' + string if line != "" and line != " " else string
            if 'EOL' in the_rest:
                #termination = "\n!\n" if self._to_exclamate(line) else "\n"
                if self._to_exclamate(line):
                    termination = "\n !\n" if level >= 1 else "\n!\n"
                else:
                    #termination = " \n" if level == 1 else "\n"
                    termination = "\n"
                rc_data['rc_str'] += line + termination
                line = ""
            self._build_line(rc_data, the_rest, line, level + 1)
            if level == 0:
                rc_data['rc_str'] += "!\n"

    def _to_exclamate(self, line):
        for statement in self.exclamation:
            if line.startswith(statement):
                return True
        return False

    def _process_next_level(self, parent, current, remaining_lines,
                            last_processed, is_root=False):
        if not remaining_lines:
            return
        pre, cmd_line = self._get_command_prepending(remaining_lines[0])
        if pre is None:
            self._process_set(cmd_line, parent, current,
                              remaining_lines, last_processed, is_root)
        elif pre.lower() == "no":
            self._process_unset(cmd_line.split(" "), current)

    def _process_set(self, cmd_line, parent, current, remaining_lines,
                     last_processed, is_root):
        cmd, args = self._get_command_and_args(cmd_line)
        if cmd in self.log_only_commands:
            self._process_next_level(parent, current,
                                     remaining_lines[1:], last_processed)
            return
        if cmd in self.parent_bound_commands:
            this_one = last_processed.get(cmd)
            start = last_processed
        else:
            this_one = parent.get(cmd)
            start = current
        if this_one is None:
            this_one, current_parent = self._get_successor_and_its_parent(
                parent, cmd, start, is_root)
        else:
            current_parent = start
        for arg in args:
            next_one, current_parent = self._get_successor_and_its_parent(
                current_parent, arg, this_one, is_root)
            this_one = next_one
        this_one['EOL'] = True
        if is_root is True:
            current = this_one
        self._process_next_level(current_parent, current, remaining_lines[1:],
                                 this_one)

    def _process_unset(self, remaining, current):
        if not remaining:
            return
        arg = remaining[0]
        rest = remaining[1:]
        if arg in current:
            if not rest:
                del current[arg]
            else:
                self._process_unset(rest, current[arg])
                num_items = len(current[arg])
                if num_items == 0:
                    del current[arg]

    def _get_successor_and_its_parent(self, parent, string, current, is_root):
        successor = current.get(string)
        if successor is None:
            successor = self._get_dict()
            current[string] = successor
            current_parent = parent if is_root is False else successor
        else:
            current_parent = current
        return successor, current_parent

    def _get_command_lines(self, snippet):
        if not snippet:
            return []
        lines = snippet.split('\n')
        commands = []
        for line in lines:
            if self._should_skip_line(line):
                continue
            cmd = self._get_embedded_command_string(line)
            if cmd is not None:
                commands.append(cmd)
        return commands

    def _should_skip_line(self, line):
        if line == "":
            return True
        if line.find("config>") != -1:
            return True
        elif line.find("cli-config-data>") != -1:
            return True
        return False

    def _get_embedded_command_string(self, line):
        match_obj = re.match(r'\s*<cmd>(.*)</cmd>\s*', line)
        if match_obj:
            return match_obj.group(1)
        return None

    def _get_command_prepending(self, cmd):
        match_obj = re.match(r'\s*(no|do) (.*)\s*', cmd)
        if match_obj:
            return match_obj.group(1), match_obj.group(2)
        return None, cmd

    def _get_command_and_args(self, cmd_line):
        str_list = cmd_line.split(" ")
        return str_list[0], str_list[1:]

    def _get_dict(self):
        return {}


# A simple Cisco IOS XE CLI simulator
class FakeRunningConfig(object):
    def __init__(self, rc):
        self._raw = rc



# Copyright (c) 2017 Cisco Systems, Inc.
# All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Basic Test Classes using RESTAPI Driver to test Cisco Nexus platforms.

These Classes are based on the original ssh event driver so same
tests occur with same configuration.  What's different between
the tests is the resulting driver output which is what
the tests in this class presents to its parent class.

You will notice in this file there are test methods which
are skipped by using 'pass'.  This is because these tests
apply to ssh only OR because rerunning the test would be
redundant.
"""

import mock
from oslo_config import cfg
import six

from networking_cisco.ml2_drivers.nexus import (
    constants as const)
from networking_cisco.ml2_drivers.nexus import (
    exceptions)
from networking_cisco.ml2_drivers.nexus import (
    nexus_db_v2 as nxos_db)
from networking_cisco.ml2_drivers.nexus import (
    nexus_restapi_snippets as snipp)
from networking_cisco.tests.unit.ml2_drivers.nexus import (
    test_cisco_nexus_base as base)
from networking_cisco.tests.unit.ml2_drivers.nexus import (
    test_cisco_nexus_events)


class TestCiscoNexusRestDeviceResults(base.TestCiscoNexusBaseResults):

    """Unit tests driver results for Cisco ML2 Nexus."""

    test_results = {
        'duplicate_add_port_driver_result': [
            [snipp.PATH_ALL,
             base.NEXUS_IP_ADDRESS_1,
             (snipp.BODY_VLAN_ADD % 267),
             base.POST],
            [(snipp.PATH_IF % 'phys-[eth1/10]'),
             base.NEXUS_IP_ADDRESS_1,
             (snipp.BODY_TRUNKVLAN % ('l1PhysIf', '', '+267')),
             base.POST]
        ],
        'duplicate_del_port_driver_result': [
            [(snipp.PATH_IF % 'phys-[eth1/10]'),
             base.NEXUS_IP_ADDRESS_1,
             (snipp.BODY_TRUNKVLAN % ('l1PhysIf', '', '-267')),
             base.POST],
            [(snipp.PATH_VLAN % '267'),
             base.NEXUS_IP_ADDRESS_1,
             '',
             base.DELETE]
        ],
        'add_port2_driver_result': [
            [snipp.PATH_ALL,
             base.NEXUS_IP_ADDRESS_1,
             (snipp.BODY_VLAN_ADD % 265),
             base.POST],
            [(snipp.PATH_IF % 'phys-[eth1/20]'),
             base.NEXUS_IP_ADDRESS_1,
             (snipp.BODY_TRUNKVLAN % ('l1PhysIf', '', '+265')),
             base.POST]
        ],
        'delete_port2_driver_result': [
            [(snipp.PATH_IF % 'phys-[eth1/20]'),
             base.NEXUS_IP_ADDRESS_1,
             (snipp.BODY_TRUNKVLAN % ('l1PhysIf', '', '-265')),
             base.POST],
            [(snipp.PATH_VLAN % '265'),
             base.NEXUS_IP_ADDRESS_1,
             '',
             base.DELETE]
        ],
        'add_port2_driver_result2': [
            [snipp.PATH_ALL,
             base.NEXUS_IP_ADDRESS_8,
             (snipp.BODY_VLAN_ADD % 267),
             base.POST],
            [(snipp.PATH_IF % 'phys-[eth1/20]'),
             base.NEXUS_IP_ADDRESS_8,
             (snipp.BODY_TRUNKVLAN % ('l1PhysIf', '', '+267')),
             base.POST]
        ],
        'delete_port2_driver_result2': [
            [(snipp.PATH_IF % 'phys-[eth1/20]'),
             base.NEXUS_IP_ADDRESS_8,
             (snipp.BODY_TRUNKVLAN % ('l1PhysIf', '', '-267')),
             base.POST]
        ],
        'add_port2_driver_result3': [
            [snipp.PATH_ALL,
             base.NEXUS_IP_ADDRESS_6,
             (snipp.BODY_VLAN_ADD % 268),
             base.POST],
            [(snipp.PATH_IF % 'aggr-[po2]'),
             base.NEXUS_IP_ADDRESS_6,
             (snipp.BODY_TRUNKVLAN % ('pcAggrIf', '', '+268')),
             base.POST],
            [snipp.PATH_ALL,
             base.NEXUS_IP_ADDRESS_7,
             (snipp.BODY_VLAN_ADD % 268),
             base.POST],
            [(snipp.PATH_IF % 'aggr-[po2]'),
             base.NEXUS_IP_ADDRESS_7,
             (snipp.BODY_TRUNKVLAN % ('pcAggrIf', '', '+268')),
             base.POST]
        ],
        'delete_port2_driver_result3': [
            [(snipp.PATH_IF % 'aggr-[po2]'),
             base.NEXUS_IP_ADDRESS_6,
             (snipp.BODY_TRUNKVLAN % ('pcAggrIf', '', '-268')),
             base.POST],
            [(snipp.PATH_VLAN % '268'),
             base.NEXUS_IP_ADDRESS_6,
             '',
             base.DELETE],
            [(snipp.PATH_IF % 'aggr-[po2]'),
             base.NEXUS_IP_ADDRESS_7,
             (snipp.BODY_TRUNKVLAN % ('pcAggrIf', '', '-268')),
             base.POST],
            [(snipp.PATH_VLAN % '268'),
             base.NEXUS_IP_ADDRESS_7,
             '',
             base.DELETE]
        ],
        'add_port_channel_driver_result': [
            [snipp.PATH_ALL,
             base.NEXUS_IP_ADDRESS_2,
             (snipp.BODY_VLAN_ADD % 268),
             base.POST],
            [(snipp.PATH_IF % 'aggr-[po2]'),
             base.NEXUS_IP_ADDRESS_2,
             (snipp.BODY_TRUNKVLAN % ('pcAggrIf', '', '+268')),
             base.POST]
        ],
        'delete_port_channel_driver_result': [
            [(snipp.PATH_IF % 'aggr-[po2]'),
             base.NEXUS_IP_ADDRESS_2,
             (snipp.BODY_TRUNKVLAN % ('pcAggrIf', '', '-268')),
             base.POST],
            [(snipp.PATH_VLAN % '268'),
             base.NEXUS_IP_ADDRESS_2,
             '',
             base.DELETE]
        ],
        'dual_add_port_driver_result': [
            [snipp.PATH_ALL,
             base.NEXUS_IP_ADDRESS_DUAL,
             (snipp.BODY_VLAN_ADD % 269),
             base.POST],
            [(snipp.PATH_IF % 'phys-[eth1/3]'),
             base.NEXUS_IP_ADDRESS_DUAL,
             (snipp.BODY_TRUNKVLAN % ('l1PhysIf', '', '+269')),
             base.POST],
            [snipp.PATH_ALL,
             base.NEXUS_IP_ADDRESS_DUAL,
             (snipp.BODY_VLAN_ADD % 269),
             base.POST],
            [(snipp.PATH_IF % 'aggr-[po2]'),
             base.NEXUS_IP_ADDRESS_DUAL,
             (snipp.BODY_TRUNKVLAN % ('pcAggrIf', '', '+269')),
             base.POST]
        ],
        'dual_delete_port_driver_result': [
            [(snipp.PATH_IF % 'phys-[eth1/3]'),
             base.NEXUS_IP_ADDRESS_DUAL,
             (snipp.BODY_TRUNKVLAN % ('l1PhysIf', '', '-269')),
             base.POST],
            [(snipp.PATH_VLAN % '269'),
             base.NEXUS_IP_ADDRESS_DUAL,
             '',
             base.DELETE],
            [(snipp.PATH_IF % 'aggr-[po2]'),
             base.NEXUS_IP_ADDRESS_DUAL,
             (snipp.BODY_TRUNKVLAN % ('pcAggrIf', '', '-269')),
             base.POST],
        ],
        'add_port_driver_result': [
            [snipp.PATH_ALL,
             base.NEXUS_IP_ADDRESS_8,
             (snipp.BODY_VLAN_ADD % 267),
             base.POST],
            [(snipp.PATH_IF % 'phys-[eth1/10]'),
             base.NEXUS_IP_ADDRESS_8,
             (snipp.BODY_TRUNKVLAN % ('l1PhysIf', '', '+267')),
             base.POST]
        ],
        'del_port_driver_result': [
            [(snipp.PATH_IF % 'phys-[eth1/10]'),
             base.NEXUS_IP_ADDRESS_8,
             (snipp.BODY_TRUNKVLAN % ('l1PhysIf', '', '-267')),
             base.POST],
            [(snipp.PATH_VLAN % '267'),
             base.NEXUS_IP_ADDRESS_8,
             '',
             base.DELETE]
        ],
        'migrate_add_host2_driver_result': [
            [snipp.PATH_ALL,
             base.NEXUS_IP_ADDRESS_3,
             (snipp.BODY_VLAN_ADD % 267),
             base.POST],
            [(snipp.PATH_IF % 'phys-[eth1/20]'),
             base.NEXUS_IP_ADDRESS_3,
             (snipp.BODY_TRUNKVLAN % ('l1PhysIf', '', '+267')),
             base.POST]
        ],
    }


class TestCiscoNexusRestDevice(test_cisco_nexus_events.TestCiscoNexusDevice):

    """Unit tests for Cisco ML2 Nexus restapi device driver"""

    def setUp(self):

        cfg.CONF.set_override('switch_heartbeat_time', 0, 'ml2_cisco')

        # Call Grandfather's setUp(); otherwise parent will set driver to
        # 'ncclient' instead of 'restapi'.
        super(test_cisco_nexus_events.TestCiscoNexusDevice, self).setUp()
        self.mock_ncclient.reset_mock()
        self.results = TestCiscoNexusRestDeviceResults()

    def test_create_delete_duplicate_ports(self):
        (super(TestCiscoNexusRestDevice, self).
            test_create_delete_duplicate_ports())

    def test_create_delete_duplicate_port_transaction(self):
        (super(TestCiscoNexusRestDevice, self).
            test_create_delete_duplicate_port_transaction())

    def test_create_delete_same_switch_diff_hosts_diff_vlan(self):
        (super(TestCiscoNexusRestDevice, self).
            test_create_delete_same_switch_diff_hosts_diff_vlan())

    def test_create_delete_same_switch_diff_hosts_same_vlan(self):
        (super(TestCiscoNexusRestDevice, self).
            test_create_delete_same_switch_diff_hosts_same_vlan())

    def test_create_delete_diff_switch_same_host(self):
        (super(TestCiscoNexusRestDevice, self).
            test_create_delete_diff_switch_same_host())

    def test_create_delete_portchannel(self):
        super(TestCiscoNexusRestDevice, self).test_create_delete_portchannel()

    def test_create_delete_dual(self):
        super(TestCiscoNexusRestDevice, self).test_create_delete_dual()

    def test_create_delete_dhcp(self):
        super(TestCiscoNexusRestDevice, self).test_create_delete_dhcp()

    def test_create_delete_router_ha_intf(self):
        (super(TestCiscoNexusRestDevice, self).
            test_create_delete_router_ha_intf())

    def test_nexus_vm_migration(self):
        super(TestCiscoNexusRestDevice, self).test_nexus_vm_migration()

    def test_update_postcommit_port_not_found(self):

        port_config = self.test_configs['test_config2']
        port_context = self._generate_port_context(
            port_config, override_netid=None)

        # An exception should not be raised during update.
        self._cisco_mech_driver.update_port_postcommit(port_context)

        # Nothing should have been sent
        self._verify_results([])

        # No port bindings should exist
        self.assertRaises(exceptions.NexusPortBindingNotFound,
            nxos_db.get_nexusport_switch_bindings,
                base.NEXUS_IP_ADDRESS_1)


class TestCiscoNexusRestInitResults(base.TestCiscoNexusBaseResults):

    """Unit tests driver results for Cisco ML2 Nexus."""

    test_results = {
        # set 1 - switch 1.1.1.1 sets eth 1/10 & 1/20 to None
        # set 2 - switch 8.8.8.8 sets eth 1/10 & 1/20 to None
        # set 3 - switch 4.4.4.4 sets eth 1/3 & portchannel 2 to None
        # set 4 - switch 2.2.2.2 sets portchannel 2 to None
        # set 5 - switch 6.6.6.6 sets portchannel 2 to None
        # set 6 - switch 7.7.7.7 sets portchannel 2 to None
        'duplicate_init_port_driver_result1': [
            [(snipp.PATH_IF % 'phys-[eth1/10]'),
             base.NEXUS_IP_ADDRESS_1,
             (snipp.BODY_TRUNKVLAN % ('l1PhysIf',
                 snipp.BODY_PORT_CH_MODE, '')),
             base.POST],

            [(snipp.PATH_IF % 'phys-[eth1/20]'),
             base.NEXUS_IP_ADDRESS_1,
             (snipp.BODY_TRUNKVLAN % ('l1PhysIf',
                 snipp.BODY_PORT_CH_MODE, '')),
             base.POST],

            [(snipp.PATH_IF % 'aggr-[po2]'),
             base.NEXUS_IP_ADDRESS_2,
             (snipp.BODY_TRUNKVLAN % ('pcAggrIf',
                 snipp.BODY_PORT_CH_MODE, '')),
             base.POST],

            [(snipp.PATH_IF % 'phys-[eth1/20]'),
             base.NEXUS_IP_ADDRESS_3,
             (snipp.BODY_TRUNKVLAN % ('l1PhysIf',
                 snipp.BODY_PORT_CH_MODE, '')),
             base.POST],

            [(snipp.PATH_IF % 'phys-[eth1/3]'),
             base.NEXUS_IP_ADDRESS_DUAL,
             (snipp.BODY_TRUNKVLAN % ('l1PhysIf',
                 snipp.BODY_PORT_CH_MODE, '')),
             base.POST],

            [(snipp.PATH_IF % 'aggr-[po2]'),
             base.NEXUS_IP_ADDRESS_DUAL,
             (snipp.BODY_TRUNKVLAN % ('pcAggrIf',
                 snipp.BODY_PORT_CH_MODE, '')),
             base.POST],

            [(snipp.PATH_IF % 'aggr-[po2]'),
             base.NEXUS_IP_ADDRESS_6,
             (snipp.BODY_TRUNKVLAN % ('pcAggrIf',
                 snipp.BODY_PORT_CH_MODE, '')),
             base.POST],

            [(snipp.PATH_IF % 'aggr-[po2]'),
             base.NEXUS_IP_ADDRESS_7,
             (snipp.BODY_TRUNKVLAN % ('pcAggrIf',
                 snipp.BODY_PORT_CH_MODE, '')),
             base.POST],

            [(snipp.PATH_IF % 'phys-[eth1/10]'),
             base.NEXUS_IP_ADDRESS_8,
             (snipp.BODY_TRUNKVLAN % ('l1PhysIf',
                 snipp.BODY_PORT_CH_MODE, '')),
             base.POST],

            [(snipp.PATH_IF % 'phys-[eth1/20]'),
             base.NEXUS_IP_ADDRESS_8,
             (snipp.BODY_TRUNKVLAN % ('l1PhysIf',
                 snipp.BODY_PORT_CH_MODE, '')),
             base.POST],

        ],

    }

GET_INTERFACE_NO_TRUNK_RESPONSE = {
    "totalCount": "1",
    "imdata": [
        {
            "l1PhysIf": {
                "attributes": {
                    "trunkVlans": "1-4094"
                }
            }
        }
    ]
}

GET_INTERFACE_PCHAN_NO_TRUNK_RESPONSE = {
    "totalCount": "1",
    "imdata": [
        {
            "pcAggrIf": {
                "attributes": {
                    "trunkVlans": "1-4094"
                }
            }
        }
    ]
}


# Skipped inheriting event class TestCiscoNexusDeviceFailure
# since some tests are generic and need not be executed twice
# and some apply only to SSH driver.


class TestCiscoNexusRestDeviceInit(
    test_cisco_nexus_events.TestCiscoNexusDeviceInit):
    """Verifies interface vlan allowed none is set when missing."""

    def get_init_side_effect(
        self, action, ipaddr=None, body=None, headers=None):

        eth_path = 'api/mo/sys/intf/phys-'
        port_chan_path = 'api/mo/sys/intf/aggr-'

        if action == snipp.PATH_GET_NEXUS_TYPE:
            return base.GET_NEXUS_TYPE_RESPONSE
        elif action in snipp.PATH_GET_PC_MEMBERS:
            return base.GET_NO_PORT_CH_RESPONSE
        elif eth_path in action:
            return GET_INTERFACE_NO_TRUNK_RESPONSE
        elif port_chan_path in action:
            return GET_INTERFACE_PCHAN_NO_TRUNK_RESPONSE

        return {}

    def restapi_mock_init(self):

        # this is to prevent interface initialization from occurring
        # which adds unnecessary noise to the results.

        data_json = {'rest_get.side_effect':
                    self.get_init_side_effect}
        self.mock_ncclient.configure_mock(**data_json)

    def setUp(self):
        """Sets up mock ncclient, and switch and credentials dictionaries."""

        cfg.CONF.set_override('switch_heartbeat_time', 0, 'ml2_cisco')

        # Call Grandfather's setUp(); otherwise parent will set driver to
        # 'ncclient' instead of 'restapi'.
        super(test_cisco_nexus_events.TestCiscoNexusDeviceInit, self).setUp()
        self.results = TestCiscoNexusRestInitResults()

    def test_verify_initialization(self):
        self._verify_results(
            self.results.get_test_results(
                'duplicate_init_port_driver_result1'))


class TestCiscoNexusRestBaremetalResults(base.TestCiscoNexusBaseResults):

    """Unit tests driver results for Cisco ML2 Nexus."""

    test_results = {

        'add_port_ethernet_driver_result': [
            [snipp.PATH_ALL,
             base.NEXUS_IP_ADDRESS_1,
             (snipp.BODY_VLAN_ADD % 267),
             base.POST],
            [(snipp.PATH_IF % 'phys-[eth1/10]'),
             base.NEXUS_IP_ADDRESS_1,
             (snipp.BODY_NATIVE_TRUNKVLAN % (
                 'l1PhysIf', '', '+267', 'vlan-267')),
             base.POST]
        ],

        'delete_port_ethernet_driver_result': [
            [(snipp.PATH_IF % 'phys-[eth1/10]'),
             base.NEXUS_IP_ADDRESS_1,
             (snipp.BODY_NATIVE_TRUNKVLAN % ('l1PhysIf', '', '-267', '')),
             base.POST],
            [(snipp.PATH_VLAN % '267'),
             base.NEXUS_IP_ADDRESS_1,
             '',
             base.DELETE]
        ],

        'add_vm_port_ethernet_driver_result': [
            [snipp.PATH_ALL,
             base.NEXUS_IP_ADDRESS_1,
             (snipp.BODY_VLAN_ADD % 265),
             base.POST],
            [(snipp.PATH_IF % 'phys-[eth1/10]'),
             base.NEXUS_IP_ADDRESS_1,
             (snipp.BODY_TRUNKVLAN % ('l1PhysIf', '', '+265')),
             base.POST]
        ],

        'delete_vm_port_ethernet_driver_result': [
            [(snipp.PATH_IF % 'phys-[eth1/10]'),
             base.NEXUS_IP_ADDRESS_1,
             (snipp.BODY_TRUNKVLAN % ('l1PhysIf', '', '-265')),
             base.POST],
            [(snipp.PATH_VLAN % '265'),
             base.NEXUS_IP_ADDRESS_1,
             '',
             base.DELETE]
        ],

        'add_port_channel_driver_result': [
            [snipp.PATH_ALL,
             base.NEXUS_IP_ADDRESS_1,
             (snipp.BODY_VLAN_ADD % 267),
             base.POST],
            [(snipp.PATH_IF % 'aggr-[po469]'),
             base.NEXUS_IP_ADDRESS_1,
             (snipp.BODY_TRUNKVLAN % ('pcAggrIf', '', '+267')),
             base.POST]
        ],

        'delete_port_channel_driver_result': [
            [(snipp.PATH_IF % 'aggr-[po469]'),
             base.NEXUS_IP_ADDRESS_1,
             (snipp.BODY_TRUNKVLAN % ('pcAggrIf', '', '-267')),
             base.POST],
            [(snipp.PATH_VLAN % '267'),
             base.NEXUS_IP_ADDRESS_1,
             '',
             base.DELETE]
        ],

        'add_port_ethernet_native_driver_result': [
            [snipp.PATH_ALL,
             base.NEXUS_IP_ADDRESS_1,
             (snipp.BODY_VLAN_ADD % 265),
             base.POST],
            [(snipp.PATH_IF % 'phys-[eth1/10]'),
             base.NEXUS_IP_ADDRESS_1,
             (snipp.BODY_NATIVE_TRUNKVLAN % (
                 'l1PhysIf', '', '+265', 'vlan-265')),
             base.POST]
        ],

        'delete_port_ethernet_native_driver_result': [
            [(snipp.PATH_IF % 'phys-[eth1/10]'),
             base.NEXUS_IP_ADDRESS_1,
             (snipp.BODY_NATIVE_TRUNKVLAN % ('l1PhysIf', '', '-265', '')),
             base.POST],
            [(snipp.PATH_VLAN % '265'),
             base.NEXUS_IP_ADDRESS_1,
             '',
             base.DELETE]
        ],

        'driver_result_unique_vPC_add1': [
            [snipp.PATH_ALL,
             base.NEXUS_IP_ADDRESS_1,
             (snipp.BODY_VLAN_ADD % 267),
             base.POST],
            [(snipp.PATH_IF % 'aggr-[po469]'),
             base.NEXUS_IP_ADDRESS_1,
             (snipp.BODY_NATIVE_TRUNKVLAN % (
                 'pcAggrIf', '', '+267', 'vlan-267')),
             base.POST],
            [snipp.PATH_ALL,
             base.NEXUS_IP_ADDRESS_2,
             (snipp.BODY_VLAN_ADD % 267),
             base.POST],
            [(snipp.PATH_IF % 'aggr-[po469]'),
             base.NEXUS_IP_ADDRESS_2,
             (snipp.BODY_NATIVE_TRUNKVLAN % (
                 'pcAggrIf', '', '+267', 'vlan-267')),
             base.POST]
        ],

        'driver_result_unique_vPC_del1': [
            [(snipp.PATH_IF % 'aggr-[po469]'),
             base.NEXUS_IP_ADDRESS_1,
             (snipp.BODY_NATIVE_TRUNKVLAN % ('pcAggrIf', '', '-267', '')),
             base.POST],
            [(snipp.PATH_VLAN % '267'),
             base.NEXUS_IP_ADDRESS_1,
             '',
             base.DELETE],
            [(snipp.PATH_IF % 'aggr-[po469]'),
             base.NEXUS_IP_ADDRESS_2,
             (snipp.BODY_NATIVE_TRUNKVLAN % ('pcAggrIf', '', '-267', '')),
             base.POST],
            [(snipp.PATH_VLAN % '267'),
             base.NEXUS_IP_ADDRESS_2,
             '',
             base.DELETE]
        ],


        'driver_result_unique_vPC_add1_vm': [
            [snipp.PATH_ALL,
             base.NEXUS_IP_ADDRESS_1,
             (snipp.BODY_VLAN_ADD % 265),
             base.POST],
            [(snipp.PATH_IF % 'aggr-[po469]'),
             base.NEXUS_IP_ADDRESS_1,
             (snipp.BODY_TRUNKVLAN % ('pcAggrIf', '', '+265')),
             base.POST],
            [snipp.PATH_ALL,
             base.NEXUS_IP_ADDRESS_2,
             (snipp.BODY_VLAN_ADD % 265),
             base.POST],
            [(snipp.PATH_IF % 'aggr-[po469]'),
             base.NEXUS_IP_ADDRESS_2,
             (snipp.BODY_TRUNKVLAN % ('pcAggrIf', '', '+265')),
             base.POST]
        ],

        'driver_result_unique_vPC_del1_vm': [
            [(snipp.PATH_IF % 'aggr-[po469]'),
             base.NEXUS_IP_ADDRESS_1,
             (snipp.BODY_TRUNKVLAN % ('pcAggrIf', '', '-265')),
             base.POST],
            [(snipp.PATH_VLAN % '265'),
             base.NEXUS_IP_ADDRESS_1,
             '',
             base.DELETE],
            [(snipp.PATH_IF % 'aggr-[po469]'),
             base.NEXUS_IP_ADDRESS_2,
             (snipp.BODY_TRUNKVLAN % ('pcAggrIf', '', '-265')),
             base.POST],
            [(snipp.PATH_VLAN % '265'),
             base.NEXUS_IP_ADDRESS_2,
             '',
             base.DELETE]
        ],

        'driver_result_unique_auto_vPC_vm_add1': [
            [snipp.PATH_ALL,
             base.NEXUS_IP_ADDRESS_1,
             (snipp.BODY_VLAN_ADD % 265),
             base.POST],
            [(snipp.PATH_IF % 'aggr-[po1001]'),
             base.NEXUS_IP_ADDRESS_1,
             (snipp.BODY_TRUNKVLAN % ('pcAggrIf', '', '+265')),
             base.POST],
            [snipp.PATH_ALL,
             base.NEXUS_IP_ADDRESS_2,
             (snipp.BODY_VLAN_ADD % 265),
             base.POST],
            [(snipp.PATH_IF % 'aggr-[po1001]'),
             base.NEXUS_IP_ADDRESS_2,
             (snipp.BODY_TRUNKVLAN % ('pcAggrIf', '', '+265')),
             base.POST]
        ],

        'driver_result_unique_auto_vPC_vm_del1': [
            [(snipp.PATH_IF % 'aggr-[po1001]'),
             base.NEXUS_IP_ADDRESS_1,
             (snipp.BODY_TRUNKVLAN % ('pcAggrIf', '', '-265')),
             base.POST],
            [(snipp.PATH_VLAN % '265'),
             base.NEXUS_IP_ADDRESS_1,
             '',
             base.DELETE],
            [(snipp.PATH_IF % 'aggr-[po1001]'),
             base.NEXUS_IP_ADDRESS_2,
             (snipp.BODY_TRUNKVLAN % ('pcAggrIf', '', '-265')),
             base.POST],
            [(snipp.PATH_VLAN % '265'),
             base.NEXUS_IP_ADDRESS_2,
             '',
             base.DELETE]
        ],

        'driver_result_unique_auto_vPC_add1': [
            [snipp.PATH_ALL,
             base.NEXUS_IP_ADDRESS_1,
             (snipp.BODY_ADD_PORT_CH % (1001, 1001, 1001)),
             base.POST],
            [snipp.PATH_ALL,
             base.NEXUS_IP_ADDRESS_1,
             (snipp.BODY_ADD_PORT_CH_P2 % (1001, 1001)),
             base.POST],
            [snipp.PATH_ALL,
             base.NEXUS_IP_ADDRESS_1,
             (snipp.BODY_ADD_CH_GRP % (1001, 1001, 'phys-[eth1/10]')),
             base.POST],
            [(snipp.PATH_IF % 'aggr-[po1001]'),
             base.NEXUS_IP_ADDRESS_1,
             (snipp.BODY_TRUNKVLAN % (
                 'pcAggrIf', snipp.BODY_PORT_CH_MODE, '')),
             base.POST],
            [snipp.PATH_ALL,
             base.NEXUS_IP_ADDRESS_2,
             (snipp.BODY_ADD_PORT_CH % (1001, 1001, 1001)),
             base.POST],
            [snipp.PATH_ALL,
             base.NEXUS_IP_ADDRESS_2,
             (snipp.BODY_ADD_PORT_CH_P2 % (1001, 1001)),
             base.POST],
            [snipp.PATH_ALL,
             base.NEXUS_IP_ADDRESS_2,
             (snipp.BODY_ADD_CH_GRP % (1001, 1001, 'phys-[eth1/20]')),
             base.POST],
            [(snipp.PATH_IF % 'aggr-[po1001]'),
             base.NEXUS_IP_ADDRESS_2,
             (snipp.BODY_TRUNKVLAN % (
                 'pcAggrIf', snipp.BODY_PORT_CH_MODE, '')),
             base.POST],
            [snipp.PATH_ALL,
             base.NEXUS_IP_ADDRESS_1,
             (snipp.BODY_VLAN_ADD % 267),
             base.POST],
            [(snipp.PATH_IF % 'aggr-[po1001]'),
             base.NEXUS_IP_ADDRESS_1,
             (snipp.BODY_NATIVE_TRUNKVLAN % (
                 'pcAggrIf', '', '+267', 'vlan-267')),
             base.POST],
            [snipp.PATH_ALL,
             base.NEXUS_IP_ADDRESS_2,
             (snipp.BODY_VLAN_ADD % 267),
             base.POST],
            [(snipp.PATH_IF % 'aggr-[po1001]'),
             base.NEXUS_IP_ADDRESS_2,
             (snipp.BODY_NATIVE_TRUNKVLAN % (
                 'pcAggrIf', '', '+267', 'vlan-267')),
             base.POST]
        ],
        'driver_result_unique_auto_vPC_del1': [
            [(snipp.PATH_IF % 'aggr-[po1001]'),
             base.NEXUS_IP_ADDRESS_1,
             (snipp.BODY_NATIVE_TRUNKVLAN % ('pcAggrIf', '', '-267', '')),
             base.POST],
            [(snipp.PATH_VLAN % '267'),
             base.NEXUS_IP_ADDRESS_1,
             '',
             base.DELETE],
            [snipp.PATH_ALL,
             base.NEXUS_IP_ADDRESS_1,
             (snipp.BODY_DEL_CH_GRP % ('1001', 'phys-[eth1/10]')),
             base.POST],
            [snipp.PATH_ALL,
             base.NEXUS_IP_ADDRESS_1,
             (snipp.BODY_DEL_PORT_CH % ('1001')),
             base.POST],
            [(snipp.PATH_IF % 'aggr-[po1001]'),
             base.NEXUS_IP_ADDRESS_2,
             (snipp.BODY_NATIVE_TRUNKVLAN % ('pcAggrIf', '', '-267', '')),
             base.POST],
            [(snipp.PATH_VLAN % '267'),
             base.NEXUS_IP_ADDRESS_2,
             '',
             base.DELETE],
            [snipp.PATH_ALL,
             base.NEXUS_IP_ADDRESS_2,
             (snipp.BODY_DEL_CH_GRP % ('1001', 'phys-[eth1/20]')),
             base.POST],
            [snipp.PATH_ALL,
             base.NEXUS_IP_ADDRESS_2,
             (snipp.BODY_DEL_PORT_CH % ('1001')),
             base.POST]
        ],
        'driver_result_unique_auto_vPC_inconsistency_failure': [
            [snipp.PATH_ALL,
             base.NEXUS_IP_ADDRESS_1,
             (snipp.BODY_ADD_PORT_CH % (1001, 1001, 1001)),
             base.POST],
            [snipp.PATH_ALL,
             base.NEXUS_IP_ADDRESS_1,
             (snipp.BODY_ADD_PORT_CH_P2 % (1001, 1001)),
             base.POST],
            [snipp.PATH_ALL,
             base.NEXUS_IP_ADDRESS_1,
             (snipp.BODY_ADD_CH_GRP % (1001, 1001, 'phys-[eth1/10]')),
             base.POST],
            [(snipp.PATH_IF % 'aggr-[po1001]'),
             base.NEXUS_IP_ADDRESS_1,
             (snipp.BODY_TRUNKVLAN % (
                 'pcAggrIf', snipp.BODY_PORT_CH_MODE, '')),
             base.POST],
            [snipp.PATH_ALL,
             base.NEXUS_IP_ADDRESS_1,
             (snipp.BODY_DEL_CH_GRP % ('1001', 'phys-[eth1/10]')),
             base.POST],
            [snipp.PATH_ALL,
             base.NEXUS_IP_ADDRESS_1,
             (snipp.BODY_DEL_PORT_CH % ('1001')),
             base.POST]
        ],
        'driver_result_unique_auto_vPC_add_usr_cmd_rest': [
            [snipp.PATH_ALL,
             base.NEXUS_IP_ADDRESS_1,
             (snipp.BODY_ADD_PORT_CH % (1001, 1001, 1001)),
             base.POST],
            [snipp.PATH_ALL,
             base.NEXUS_IP_ADDRESS_1,
             (snipp.BODY_ADD_CH_GRP % (1001, 1001, 'phys-[eth1/10]')),
             base.POST],
            [(snipp.PATH_IF % 'aggr-[po1001]'),
             base.NEXUS_IP_ADDRESS_1,
             (snipp.BODY_TRUNKVLAN % (
                 'pcAggrIf', snipp.BODY_PORT_CH_MODE, '')),
             base.POST],
            [snipp.PATH_ALL,
             base.NEXUS_IP_ADDRESS_2,
             (snipp.BODY_ADD_PORT_CH % (1001, 1001, 1001)),
             base.POST],
            [snipp.PATH_ALL,
             base.NEXUS_IP_ADDRESS_2,
             (snipp.BODY_ADD_CH_GRP % (1001, 1001, 'phys-[eth1/20]')),
             base.POST],
            [(snipp.PATH_IF % 'aggr-[po1001]'),
             base.NEXUS_IP_ADDRESS_2,
             (snipp.BODY_TRUNKVLAN % (
                 'pcAggrIf', snipp.BODY_PORT_CH_MODE, '')),
             base.POST],
            [snipp.PATH_ALL,
             base.NEXUS_IP_ADDRESS_1,
             (snipp.BODY_VLAN_ADD % 267),
             base.POST],
            [(snipp.PATH_IF % 'aggr-[po1001]'),
             base.NEXUS_IP_ADDRESS_1,
             (snipp.BODY_NATIVE_TRUNKVLAN % (
                 'pcAggrIf', '', '+267', 'vlan-267')),
             base.POST],
            [snipp.PATH_ALL,
             base.NEXUS_IP_ADDRESS_2,
             (snipp.BODY_VLAN_ADD % 267),
             base.POST],
            [(snipp.PATH_IF % 'aggr-[po1001]'),
             base.NEXUS_IP_ADDRESS_2,
             (snipp.BODY_NATIVE_TRUNKVLAN % (
                 'pcAggrIf', '', '+267', 'vlan-267')),
             base.POST],
        ],

        'driver_result_unique_auto_vPC_add_usr_cmd_nxapi_cli': [
            [snipp.PATH_USER_CMDS,
             base.NEXUS_IP_ADDRESS_1,
             "int port-channel 1001 ;spanning-tree port type edge trunk "
             ";no lacp suspend-individual",
             base.POST],
            [snipp.PATH_USER_CMDS,
             base.NEXUS_IP_ADDRESS_2,
             "int port-channel 1001 ;spanning-tree port type edge trunk "
             ";no lacp suspend-individual",
             base.POST],
        ],

    }

GET_PORT_CH_RESPONSE = {
    "totalCount": "4",
    "imdata": [
        {
            "pcRsMbrIfs": {
                "attributes": {
                    "parentSKey": "po1",
                    "tSKey": "eth1/11",
                }
            }
        },
        {
            "pcRsMbrIfs": {
                "attributes": {
                    "parentSKey": "po469",
                    "tSKey": "eth1/10",
                }
            }
        },
        {
            "pcRsMbrIfs": {
                "attributes": {
                    "parentSKey": "po2",
                    "tSKey": "eth1/12",
                }
            }
        },
        {
            "pcRsMbrIfs": {
                "attributes": {
                    "parentSKey": "po470",
                    "tSKey": "eth1/20",
                }
            }
        }
    ]
}


class TestCiscoNexusRestBaremetalDevice(
    test_cisco_nexus_events.TestCiscoNexusBaremetalDevice):

    """Tests for Cisco ML2 Nexus baremetal RESTAPI device driver."""

    def get_init_side_effect(
        self, action, ipaddr=None, body=None, headers=None):

        eth_path = 'api/mo/sys/intf/phys-'
        port_chan_path = 'api/mo/sys/intf/aggr-'

        if action == snipp.PATH_GET_NEXUS_TYPE:
            return base.GET_NEXUS_TYPE_RESPONSE
        elif action in snipp.PATH_GET_PC_MEMBERS:
            return GET_PORT_CH_RESPONSE
        elif eth_path in action:
            return base.GET_INTERFACE_RESPONSE
        elif port_chan_path in action:
            return base.GET_INTERFACE_PCHAN_RESPONSE

        return {}

    def get_init_side_effect2(
        self, action, ipaddr=None, body=None, headers=None):

        eth_path = 'api/mo/sys/intf/phys-'
        port_chan_path = 'api/mo/sys/intf/aggr-'

        if action == snipp.PATH_GET_NEXUS_TYPE:
            return base.GET_NEXUS_TYPE_RESPONSE
        elif action in snipp.PATH_GET_PC_MEMBERS:
            return base.GET_NO_PORT_CH_RESPONSE
        elif eth_path in action:
            return base.GET_INTERFACE_RESPONSE
        elif port_chan_path in action:
            return GET_INTERFACE_PCHAN_NO_TRUNK_RESPONSE

        return {}

    def _init_port_channel(self, which=1):

        # this is to prevent interface initialization from occurring
        # which adds unnecessary noise to the results.

        GET_PORT_CH_RESPONSE['imdata'][which]['pcRsMbrIfs'][
            'attributes']['parentSKey'] = 'po469'
        data_json = {'rest_get.side_effect':
                    self.get_init_side_effect}
        self.mock_ncclient.configure_mock(**data_json)

    def setUp(self):
        """Sets up mock ncclient, and switch and credentials dictionaries."""
        original_intersect = nxos_db._get_free_vpcids_on_switches

        def new_get_free_vpcids_on_switches(nexus_ips):
            intersect = list(original_intersect(nexus_ips))
            intersect.sort()
            return intersect

        mock.patch.object(nxos_db,
                         '_get_free_vpcids_on_switches',
                         new=new_get_free_vpcids_on_switches).start()

        cfg.CONF.set_override('switch_heartbeat_time', 0, 'ml2_cisco')

        # Call Grandfather's setUp(); otherwise parent will set driver to
        # 'ncclient' instead of 'restapi'.
        super(test_cisco_nexus_events.TestCiscoNexusBaremetalDevice,
              self).setUp()
        self.results = TestCiscoNexusRestBaremetalResults()

    def test_create_delete_basic_bm_ethernet_port_and_vm(self):
        (super(TestCiscoNexusRestBaremetalDevice, self).
            test_create_delete_basic_bm_ethernet_port_and_vm())

    def test_create_delete_basic_port_channel(self):
        """Basic creation and deletion test of 1 learned port-channel."""
        (super(TestCiscoNexusRestBaremetalDevice, self).
            test_create_delete_basic_port_channel())

    def test_create_delete_learn_vpc_and_vm(self):
        (super(TestCiscoNexusRestBaremetalDevice, self).
            test_create_delete_learn_vpc_and_vm())

    def test_create_delete_basic_eth_port_is_native(self):
        (super(TestCiscoNexusRestBaremetalDevice, self).
            test_create_delete_basic_eth_port_is_native())

    def test_create_delete_switch_ip_not_defined(self):
        (super(TestCiscoNexusRestBaremetalDevice, self).
            test_create_delete_switch_ip_not_defined())

    def test_automated_port_channel_creation_deletion(self):
        """Basic creation and deletion test of 1 auto port-channel."""

        data_json = {'rest_get.side_effect':
                    self.get_init_side_effect2}
        self.mock_ncclient.configure_mock(**data_json)

        switch_list = ['1.1.1.1', '2.2.2.2']
        for switch_ip in switch_list:
            cfg.CONF.set_override(
                const.VPCPOOL, ('1001-1025, 1030'),
                cfg.CONF.ml2_cisco.nexus_switches.get(switch_ip)._group)
        self._cisco_mech_driver._initialize_vpc_alloc_pools()

        self._basic_create_verify_port_vlan(
            'test_config_vPC',
            self.results.get_test_results(
                'driver_result_unique_auto_vPC_add1'),
            nbr_of_bindings=2)

        for switch_ip in switch_list:
            self.assertEqual(
                25, len(nxos_db.get_free_switch_vpc_allocs(switch_ip)))

        # Clean all the ncclient mock_calls so we can evaluate
        # results of delete operations.
        self.mock_ncclient.reset_mock()

        self._basic_delete_verify_port_vlan(
            'test_config_vPC',
            self.results.get_test_results(
                'driver_result_unique_auto_vPC_del1'))

        for switch_ip in switch_list:
            self.assertEqual(
                26, len(nxos_db.get_free_switch_vpc_allocs(switch_ip)))

    def test_create_delete_automated_vpc_and_vm(self):
        """Basic creation and deletion test of 2 auto port-channel and vm."""

        data_json = {'rest_get.side_effect':
                    self.get_init_side_effect2}
        self.mock_ncclient.configure_mock(**data_json)

        switch_list = ['1.1.1.1', '2.2.2.2']
        for switch_ip in switch_list:
            cfg.CONF.set_override(
                const.VPCPOOL, ('1001-1025, 1030'),
                cfg.CONF.ml2_cisco.nexus_switches.get(switch_ip)._group)
        self._cisco_mech_driver._initialize_vpc_alloc_pools()

        self._basic_create_verify_port_vlan(
            'test_config_vPC',
            self.results.get_test_results(
                'driver_result_unique_auto_vPC_add1'),
            nbr_of_bindings=2)

        # Clean all the ncclient mock_calls so we can evaluate
        # results of delete operations.
        self.mock_ncclient.reset_mock()

        self._basic_create_verify_port_vlan(
            'test_config_vm',
            self.results.get_test_results(
                'driver_result_unique_auto_vPC_vm_add1'),
            nbr_of_bindings=4)

        for switch_ip in switch_list:
            self.assertEqual(
                25, len(nxos_db.get_free_switch_vpc_allocs(switch_ip)))

        self._basic_delete_verify_port_vlan(
            'test_config_vm',
            self.results.get_test_results(
                'driver_result_unique_auto_vPC_vm_del1'),
            nbr_of_bindings=2)

        self._basic_delete_verify_port_vlan(
            'test_config_vPC',
            self.results.get_test_results(
                'driver_result_unique_auto_vPC_del1'))

        for switch_ip in switch_list:
            self.assertEqual(
                26, len(nxos_db.get_free_switch_vpc_allocs(switch_ip)))

    def test_automated_port_channel_w_user_cfg(self):
        """Basic creation and deletion test of 1 auto port-channel."""

        data_json = {'rest_get.side_effect':
                    self.get_init_side_effect2}
        self.mock_ncclient.configure_mock(**data_json)

        switch_list = ['1.1.1.1', '2.2.2.2']

        for switch_ip in switch_list:
            cfg.CONF.set_override(
                const.VPCPOOL, ('1001-1025'),
                cfg.CONF.ml2_cisco.nexus_switches.get(switch_ip)._group)
        self._cisco_mech_driver._initialize_vpc_alloc_pools()

        self._cfg_vPC_user_commands(
            switch_list, "spanning-tree port type edge trunk ;no lacp "
                         "suspend-individual")

        self._basic_create_verify_port_vlan(
            'test_config_vPC',
            self.results.get_test_results(
                'driver_result_unique_auto_vPC_add_usr_cmd_rest'),
            nbr_of_bindings=2)

        self._verify_nxapi_results(
            self.results.get_test_results(
                'driver_result_unique_auto_vPC_add_usr_cmd_nxapi_cli'))

        # Clean all the ncclient mock_calls so we can evaluate
        # results of delete operations.
        self.mock_ncclient.reset_mock()

        self._basic_delete_verify_port_vlan(
            'test_config_vPC',
            self.results.get_test_results(
                'driver_result_unique_auto_vPC_del1'))

        for switch_ip in switch_list:
            self.assertEqual(
                25, len(nxos_db.get_free_switch_vpc_allocs(switch_ip)))

    def test_failure_inconsistent_learned_chgrp(self):
        """Learning chgrp but different on both eth interfaces."""

        # Clean all the ncclient mock_calls to clear exception
        # and other mock_call history.
        self.mock_ncclient.reset_mock()

        LOCAL_GET_PORT_CH_RESPONSE = {
            "totalCount": "2",
            "imdata": [
                {
                    "pcRsMbrIfs": {
                        "attributes": {
                            "parentSKey": "po469",
                            "tSKey": "eth1/10",
                        }
                    }
                },
                {
                    "pcRsMbrIfs": {
                        "attributes": {
                            "parentSKey": "po470",
                            "tSKey": "eth1/20",
                        }
                    }
                }
            ]
        }

        def local_get_init_side_effect(
            action, ipaddr=None, body=None, headers=None):

            eth_path = 'api/mo/sys/intf/phys-'
            port_chan_path = 'api/mo/sys/intf/aggr-'

            if action == snipp.PATH_GET_NEXUS_TYPE:
                return base.GET_NEXUS_TYPE_RESPONSE
            elif action in snipp.PATH_GET_PC_MEMBERS:
                return LOCAL_GET_PORT_CH_RESPONSE
            elif eth_path in action:
                return base.GET_INTERFACE_RESPONSE
            elif port_chan_path in action:
                return GET_INTERFACE_PCHAN_NO_TRUNK_RESPONSE

            return {}

        # Substitute init_port_channel() with the following
        # since this is a one time test scenario.
        data_json = {'rest_get.side_effect':
                    local_get_init_side_effect}
        self.mock_ncclient.configure_mock(**data_json)

        e = self.assertRaises(exceptions.NexusVPCLearnedNotConsistent,
                              self._create_port,
                              self.test_configs[
                                  'test_config_vPC'])
        x = six.u(str(e))
        self.assertIn("first interface 1.1.1.1, ethernet:1/10, vpc=469", x)
        self.assertIn("second interface 2.2.2.2, ethernet:1/20, vpc=470", x)

    def test_failure_inconsistent_new_chgrp(self):
        """Started as newly created chgrp but one if had chgrp configured."""

        # First interface Nexus returns there's no ch_grp
        # so treat as port-channel create.
        # Second interface Nexus returns ch_grp so so process
        # reset procedure which checks that .....
        #   - port-channel deleted from Nexus for first interface
        #   - ch_grp removed from Nexus on first interface
        #   - free-up vpcid allocated on first interface
        #   - raised cexc.NexusVPCExpectedNoChgrp

        LOCAL_GET_PORT_CH_RESPONSE = {
            "totalCount": "1",
            "imdata": [
                {
                    "pcRsMbrIfs": {
                        "attributes": {
                            "parentSKey": "po470",
                            "tSKey": "eth1/20",
                        }
                    }
                }
            ]
        }

        def local_get_init_side_effect(
            action, ipaddr=None, body=None, headers=None):

            eth_path = 'api/mo/sys/intf/phys-'
            port_chan_path = 'api/mo/sys/intf/aggr-'

            if action == snipp.PATH_GET_NEXUS_TYPE:
                return base.GET_NEXUS_TYPE_RESPONSE
            elif action in snipp.PATH_GET_PC_MEMBERS:
                return LOCAL_GET_PORT_CH_RESPONSE
            elif eth_path in action:
                return base.GET_INTERFACE_RESPONSE
            elif port_chan_path in action:
                return GET_INTERFACE_PCHAN_NO_TRUNK_RESPONSE

            return {}

        # Substitute init_port_channel() with the following
        # since this is a one time test scenario.
        data_json = {'rest_get.side_effect':
                    local_get_init_side_effect}
        self.mock_ncclient.configure_mock(**data_json)

        switch_list = ['1.1.1.1', '2.2.2.2']

        for switch_ip in switch_list:
            nxos_db.init_vpc_entries(switch_ip,
                self._make_vpc_list(1001, 1025))
            allocs = nxos_db.get_free_switch_vpc_allocs(switch_ip)
            self.assertEqual(len(allocs), 25)

        e = self.assertRaises(exceptions.NexusVPCExpectedNoChgrp,
                              self._create_port,
                              self.test_configs[
                                  'test_config_vPC'])
        # Check that appropriate string in exception string
        x = six.u(str(e))
        self.assertIn("first interface 1.1.1.1, ethernet:1/10, vpc=None", x)
        self.assertIn("second interface 2.2.2.2, ethernet:1/20, vpc=470", x)

        # Verify vpcid initially allocated is now free
        for switch_ip in switch_list:
            allocs = nxos_db.get_free_switch_vpc_allocs(switch_ip)
            self.assertEqual(len(allocs), 25)

        # Verify no attempt to create port-channels
        self._verify_results([])

    def test_vpcids_depleted_failure(self):
        """Verifies exception when failed to get vpcid."""

        # Clean all the ncclient mock_calls to clear exception
        # and other mock_call history.
        self.mock_ncclient.reset_mock()

        def new_alloc_vpcid(nexus_ip_list):
            return 0

        mock.patch.object(nxos_db,
                         'alloc_vpcid',
                          new=new_alloc_vpcid).start()
        e = self.assertRaises(exceptions.NexusVPCAllocFailure,
                              self._create_port,
                              self.test_configs[
                                  'test_config_vPC'])
        x = six.u(str(e))
        self.assertIn("switches=1.1.1.1,2.2.2.2", x)

        # Clean all the ncclient mock_calls to clear exception
        # and other mock_call history.
        self.mock_ncclient.reset_mock()


class TestCiscoNexusBaremetalVPCConfig(base.TestCiscoNexusBase,
                                       test_cisco_nexus_events.
                                       TestCiscoNexusDeviceConfig,
                                       TestCiscoNexusRestDeviceResults):

    """Unit tests for Cisco ML2 Nexus baremetal VPC Config.

    The purpose of this test case is to validate vpc pool initialization.
    If vpc-pool is configured, it will be compared with what currently
    exists in the vpc pool data base. Adds and removals of the data base
    will occur.  Removal will not occur if the entry is active.
    """

    def setUp(self):
        super(TestCiscoNexusBaremetalVPCConfig, self).setUp()
        self.mock_ncclient.reset_mock()

    def _run_vpc_config_test(self, switch_ip, config, count_in,
                             min_in, max_in):
        """Config vpc-pool config with garbage. log & no db entries."""

        cfg.CONF.set_override(
            const.VPCPOOL, config,
            cfg.CONF.ml2_cisco.nexus_switches.get(switch_ip)._group)

        self._cisco_mech_driver._initialize_vpc_alloc_pools()

        # Verify get_switch_vpc_count_min_max() returns correct
        # count, min, max values for switches.
        count, min, max = nxos_db.get_switch_vpc_count_min_max(
            switch_ip)
        self.assertEqual(count, count_in)
        self.assertEqual(min, min_in)
        self.assertEqual(max, max_in)

    def test_vpc_config_db_results_bad_config1(self):
        """Config vpc-pool config with garbage. log & no db entries."""

        self._run_vpc_config_test('1.1.1.1', 'blahblahblah', 0, None, None)

    def test_vpc_config_db_results_bad_config2(self):
        """Config vpc-pool config with bad range. log & no db entries."""

        self._run_vpc_config_test('1.1.1.1', '5-7-9,1', 0, None, None)

    def test_vpc_config_db_results_bad_config3(self):
        """Config vpc-pool config with bad digits. log & no db entries."""

        self._run_vpc_config_test('1.1.1.1', '5-abc,1', 0, None, None)

    def test_vpc_config_db_results_bad_vpc_range(self):
        """Config vpc-pool config with bad min/max values."""

        # bad_min = 0-5
        bad_min = str(const.MINVPC - 1) + '-5'
        self._run_vpc_config_test('1.1.1.1', bad_min, 0, None, None)

        # bad_max = 4096-4097
        bad_max = str(const.MAXVPC) + '-' + str(const.MAXVPC + 1)
        self._run_vpc_config_test('1.1.1.1', bad_max, 0, None, None)

    def test_vpc_config_db_results_bad_config_keep_old(self):
        """Verify on config error, existing db entries stay intact."""

        old_list = [1, 6, 8, 11]

        # Pretend these already existed and make 8 active
        nxos_db.init_vpc_entries('1.1.1.1', old_list)
        nxos_db.update_vpc_entry(['1.1.1.1'], 8, True, True)

        # valid port-channel values are 1-4096 on Nexus 9K

        # ERROR: range starts with 0
        bad_min = str(const.MINVPC - 1) + '-1001, 1002'
        self._run_vpc_config_test('1.1.1.1', bad_min, 4, 1, 11)

    def test_vpc_config_db_results_removal(self):
        """Allow user to remove config but only non-active."""

        # 1 no add, already exists
        # 6 remove not active
        # 8 no remove, ACTIVE
        # 11 no add, already exists
        old_list = [1, 6, 8, 11]

        # Pretend these already existed and make 8 active
        nxos_db.init_vpc_entries('1.1.1.1', old_list)
        nxos_db.update_vpc_entry(['1.1.1.1'], 8, True, True)

        self._run_vpc_config_test('1.1.1.1', '', 1, 8, 8)

        # Make 8 inactive and try again.
        nxos_db.update_vpc_entry(['1.1.1.1'], 8, False, False)
        self._run_vpc_config_test('1.1.1.1', '', 0, None, None)

    def test_vpc_config_db_results_good_config_not_range(self):
        """Config valid vpc-pool not range config. """

        self._run_vpc_config_test('1.1.1.1', '1,3,5', 3, 1, 5)

    def test_vpc_config_db_results_good_config_range(self):
        """Config valid vpc-pool range config. """

        self._run_vpc_config_test('1.1.1.1', '1-5', 5, 1, 5)

    def test_vpc_config_db_results_good_config_all(self):
        """Config valid vpc-pool range config. Test Min/Max vpc value."""

        # test_range_limits = 1-5,4096
        test_range_limits = str(const.MINVPC) + '-5,' + str(const.MAXVPC)
        self._run_vpc_config_test('1.1.1.1', test_range_limits,
                                  6, const.MINVPC, const.MAXVPC)

    def test_vpc_config_db_results_with_old_config1(self):
        """Config valid vpc-pool compare with pre-existing entries."""

        # 1 will be removed,
        # 3 no add, already exists
        # 4 no add, already exists
        # 11 will not be removed since active
        old_list = [1, 3, 4, 11]

        # Pretend these already existed and make 11 active
        nxos_db.init_vpc_entries('1.1.1.1', old_list)
        nxos_db.update_vpc_entry(['1.1.1.1'], 11, True, True)

        self._run_vpc_config_test('1.1.1.1', '2-5, 8', 6, 2, 11)

    def test_vpc_config_db_results_with_old_config2(self):
        """Config valid vpc-pool compare with pre-existing entries."""

        # 1 no add, already exists
        # 6 remove not active
        # 8 no remove, ACTIVE
        # 11 no add, already exists
        old_list = [1, 6, 8, 11]

        # Pretend these already existed and make 8 active
        nxos_db.init_vpc_entries('1.1.1.1', old_list)
        nxos_db.update_vpc_entry(['1.1.1.1'], 8, True, True)

        self._run_vpc_config_test('1.1.1.1', '1-4, 9, 11', 7, 1, 11)

    def test_vpc_config_db_results_with_old_config3(self):
        """Config valid vpc-pool compare with pre-existing entries."""

        # 1 no add, already exists
        # 11 no add, already exists
        old_list = [1, 6, 8, 11]

        # Pretend these already existed and make 8 active
        nxos_db.init_vpc_entries('1.1.1.1', old_list)

        self._run_vpc_config_test('1.1.1.1', '1-4, 6-9, 11', 9, 1, 11)

# Skipped inheriting event class TestCiscoNexusNonCacheSshDevice
# since it does not apply to REST API
